# Semantic Version Parsing and Comparison
#
# Copyright (C) 2023  Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.
from __future__ import annotations
import re
from enum import Flag, auto
from typing import Tuple, Optional, Dict, List, Any

# Python regex for parsing version strings from PEP 440
# https://peps.python.org/pep-0440/#appendix-b-parsing-version-strings-with-regular-expressions
VERSION_PATTERN = r"""
    v?
    (?:
        (?:(?P<epoch>[0-9]+)!)?                           # epoch
        (?P<release>[0-9]+(?:\.[0-9]+)*)                  # release segment
        (?P<pre>                                          # pre-release
            [-_\.]?
            (?P<pre_l>(a|b|c|rc|alpha|beta|pre|preview))
            [-_\.]?
            (?P<pre_n>[0-9]+)?
        )?
        (?P<post>                                         # post release
            (?:-(?P<post_n1>[0-9]+))
            |
            (?:
                [-_\.]?
                (?P<post_l>post|rev|r)
                [-_\.]?
                (?P<post_n2>[0-9]+)?
            )
        )?
        (?P<dev>                                          # dev release
            [-_\.]?
            (?P<dev_l>dev)
            [-_\.]?
            (?P<dev_n>[0-9]+)?
        )?
    )
    (?:\+(?P<local>[a-z0-9]+(?:[-_\.][a-z0-9]+)*))?       # local version
"""

GIT_VERSION_PATTERN = r"""
    (?P<tag>
        v?
        (?P<release>[0-9]+(?:\.[0-9]+)*)                  # release segment
        (?P<pre>                                          # pre-release
            [-_\.]?
            (?P<pre_l>(a|b|c|rc|alpha|beta|pre|preview))
            [-_\.]?
            (?P<pre_n>[0-9]+)?
        )?
    )
    (?:
        (?:-(?P<dev_n>[0-9]+))                            # dev count
        (?:-g(?P<hash>[a-fA-F0-9]+))?                     # abbrev hash
    )?
    (?P<dirty>-dirty)?
    (?P<inferred>-(?:inferred|shallow))?
"""

_py_version_regex = re.compile(
    r"^\s*" + VERSION_PATTERN + r"\s*$",
    re.VERBOSE | re.IGNORECASE,
)

_git_version_regex = re.compile(
    r"^\s*" + GIT_VERSION_PATTERN + r"\s*$",
    re.VERBOSE | re.IGNORECASE,
)

class ReleaseType(Flag):
    FINAL = auto()
    ALPHA = auto()
    BETA = auto()
    RELEASE_CANDIDATE = auto()
    POST = auto()
    DEV = auto()

class BaseVersion:
    def __init__(self, version: str) -> None:
        self._release: str = "?"
        self._release_type = ReleaseType(0)
        self._tag: str = "?"
        self._orig: str = version.strip()
        self._release_tup: Tuple[int, ...] = tuple()
        self._extra_tup: Tuple[int, ...] = tuple()
        self._has_dev_part: bool = False
        self._dev_count: int = 0
        self._valid_version: bool = False

    @property
    def full_version(self) -> str:
        return self._orig

    @property
    def short_version(self) -> str:
        if not self._valid_version:
            return "?"
        return f"{self._tag}-{self._dev_count}"

    @property
    def release(self) -> str:
        return self._release

    @property
    def tag(self) -> str:
        return self._tag

    @property
    def release_type(self) -> ReleaseType:
        return self._release_type

    @property
    def dev_count(self) -> int:
        return self._dev_count

    def is_pre_release(self) -> bool:
        for pr_idx in (1, 2, 3):
            if ReleaseType(1 << pr_idx) in self._release_type:
                return True
        return False

    def is_post_release(self) -> bool:
        return ReleaseType.POST in self._release_type

    def is_dev_release(self) -> bool:
        return ReleaseType.DEV in self._release_type

    def is_alpha_release(self) -> bool:
        return ReleaseType.ALPHA in self._release_type

    def is_beta_release(self) -> bool:
        return ReleaseType.BETA in self._release_type

    def is_release_candidate(self) -> bool:
        return ReleaseType.RELEASE_CANDIDATE in self._release_type

    def is_final_release(self) -> bool:
        return ReleaseType.FINAL in self._release_type

    def is_valid_version(self) -> bool:
        return self._valid_version

    def __str__(self) -> str:
        return self._orig

    def _validate(self, other: BaseVersion) -> None:
        if not self._valid_version:
            raise ValueError(
                f"Version {self._orig} is not a valid version string "
                f"for type {type(self).__name__}"
            )
        if not other._valid_version:
            raise ValueError(
                f"Version {other._orig} is not a valid version string "
                f"for type {type(self).__name__}"
            )

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, type(self)):
            raise ValueError("Invalid type for comparison")
        self._validate(__value)
        if self._release_tup != __value._release_tup:
            return False
        if self._extra_tup != __value._extra_tup:
            return False
        if self._has_dev_part != __value._has_dev_part:
            return False
        if self._dev_count != __value._dev_count:
            return False
        return True

    def __lt__(self, __value: object) -> bool:
        if not isinstance(__value, type(self)):
            raise ValueError("Invalid type for comparison")
        self._validate(__value)
        if self._release_tup != __value._release_tup:
            return self._release_tup < __value._release_tup
        if self._extra_tup != __value._extra_tup:
            return self._extra_tup < __value._extra_tup
        if self._has_dev_part != __value._has_dev_part:
            return self._has_dev_part
        return self._dev_count < __value._dev_count

    def __le__(self, __value: object) -> bool:
        if not isinstance(__value, type(self)):
            raise ValueError("Invalid type for comparison")
        self._validate(__value)
        if self._release_tup > __value._release_tup:
            return False
        if self._extra_tup > __value._extra_tup:
            return False
        if self._has_dev_part != __value._has_dev_part:
            return self._has_dev_part
        return self._dev_count <= __value._dev_count

    def __ne__(self, __value: object) -> bool:
        if not isinstance(__value, type(self)):
            raise ValueError("Invalid type for comparison")
        self._validate(__value)
        if self._release_tup != __value._release_tup:
            return True
        if self._extra_tup != __value._extra_tup:
            return True
        if self._has_dev_part != __value._has_dev_part:
            return True
        if self._dev_count != __value._dev_count:
            return True
        return False

    def __gt__(self, __value: object) -> bool:
        if not isinstance(__value, type(self)):
            raise ValueError("Invalid type for comparison")
        self._validate(__value)
        if self._release_tup != __value._release_tup:
            return self._release_tup > __value._release_tup
        if self._extra_tup != __value._extra_tup:
            return self._extra_tup > __value._extra_tup
        if self._has_dev_part != __value._has_dev_part:
            return __value._has_dev_part
        return self._dev_count > __value._dev_count

    def __ge__(self, __value: object) -> bool:
        if not isinstance(__value, type(self)):
            raise ValueError("Invalid type for comparison")
        self._validate(__value)
        if self._release_tup < __value._release_tup:
            return False
        if self._extra_tup < __value._extra_tup:
            return False
        if self._has_dev_part != __value._has_dev_part:
            return __value._has_dev_part
        return self._dev_count >= __value._dev_count


class PyVersion(BaseVersion):
    def __init__(self, version: str) -> None:
        super().__init__(version)
        version_info = self._get_version_info()
        if version_info is None:
            return
        release: Optional[str] = version_info["release"]
        if release is None:
            return
        self._valid_version = True
        self._release = release
        self._tag = f"v{release}" if self._orig[0].lower() == "v" else release
        self._release_tup = tuple(int(part) for part in release.split("."))
        self._extra_tup = (1, 0, 0)
        if version_info["pre"] is not None:
            pre_conv: Dict[str, int]
            pre_conv = dict([("a", 1), ("b", 2), ("c", 3), ("r", 3), ("p", 3)])
            lbl: str = version_info["pre_l"][0].lower()
            self._extra_tup = (0, pre_conv.get(lbl, 0), int(version_info["pre_n"] or 0))
            self._tag += version_info["pre"]
            self._release_type |= ReleaseType(1 << pre_conv.get(lbl, 1))
            if version_info["post"] is not None:
                # strange combination of a "post" pre-release.
                num = version_info["post_n1"] or version_info["post_n2"]
                self._extra_tup += (int(num or 0),)
                self._tag += version_info["post"]
                self._release_type |= ReleaseType.POST
        elif version_info["post"] is not None:
            num = version_info["post_n1"] or version_info["post_n2"]
            self._extra_tup = (2, int(num or 0), 0)
            self._tag += version_info["post"]
            self._release_type |= ReleaseType.POST
        self._has_dev_part = version_info["dev"] is not None
        if self._has_dev_part:
            self._release_type |= ReleaseType.DEV
        elif self._release_type.value == 0:
            self._release_type = ReleaseType.FINAL
        elif self._release_type.value == ReleaseType.POST.value:
            self._release_type |= ReleaseType.FINAL
        self._dev_count = int(version_info["dev_n"] or 0)
        self.local: Optional[str] = version_info["local"]

    @property
    def short_version(self) -> str:
        if not self._valid_version:
            return "?"
        if not self._dev_count:
            return self._tag
        return f"{self._tag}-{self._dev_count}"

    def _get_version_info(self) -> Optional[Dict[str, Any]]:
        ver_match = _py_version_regex.match(self._orig)
        if ver_match is None:
            return None
        return ver_match.groupdict()

    def bump_local_version(self, new_local: str) -> PyVersion:
        # current patch version needs to be bumped
        version_info = self._get_version_info()
        if version_info is None:
            return PyVersion("?")
        new_ver = version_info["release"]
        major, _, minor = new_ver.rpartition(".")
        if major and not self._dev_count:
            new_ver = f"v{major}.{int(minor) + 1}"
        if version_info["pre"] is not None:
            new_ver += version_info["pre"]
        if version_info["post"] is not None:
            new_ver += version_info["post"]
        if self._dev_count:
            new_ver += f".dev{self._dev_count + 1}"
        else:
            new_ver += ".dev1"
        new_ver += f"+{new_local}"
        return PyVersion(new_ver)

    def convert_to_git(self) -> GitVersion:
        version_info = self._get_version_info()
        if version_info is None:
            return GitVersion("?")
        git_version: Optional[str] = version_info["release"]
        if git_version is None:
            raise ValueError("Invalid version string")
        if self._orig[0].lower() == "v":
            git_version = f"v{git_version}"
        local: str = version_info["local"] or ""
        # Assume semantic versioning, convert the version string.
        if self._dev_count > 0:
            major, _, minor = git_version.rpartition(".")
            if major:
                git_version = f"v{major}.{max(int(minor) - 1, 0)}"
        if version_info["pre"] is not None:
            git_version = f"{git_version}{version_info['pre']}"
        git_version = f"{git_version}-{self._dev_count}"
        if local:
            local_parts = re.split(r"[-_\.]", local)
            if local_parts[0]:
                git_version = f"{git_version}-{local_parts[0]}"
            if len(local_parts) > 1:
                git_version = f"{git_version}-dirty"
        return GitVersion(git_version)


class GitVersion(BaseVersion):
    def __init__(self, version: str) -> None:
        super().__init__(version)
        self._is_dirty: bool = False
        self._is_inferred: bool = False
        ver_match = _git_version_regex.match(version)
        if ver_match is None:
            # Check Fallback
            fb_match = re.match(r"(?P<hash>[a-fA-F0-9]+)(?P<dirty>-dirty)?", self._orig)
            if fb_match is None:
                return
            self._tag = ""
            self._release = fb_match["hash"]
            self._is_dirty = fb_match["dirty"] is not None
            self._is_inferred = True
            return
        version_info = ver_match.groupdict()
        release: Optional[str] = version_info["release"]
        if release is None:
            return
        self._valid_version = True
        self._release = release
        self._tag = version_info["tag"] or "?"
        self._release_tup = tuple(int(part) for part in release.split("."))
        self._extra_tup = (1, 0, 0)
        if version_info["pre"] is not None:
            pre_conv: Dict[str, int]
            pre_conv = dict([("a", 1), ("b", 2), ("c", 3), ("r", 3), ("p", 3)])
            lbl = version_info["pre_l"][0].lower()
            self._extra_tup = (0, pre_conv.get(lbl, 0), int(version_info["pre_n"] or 0))
            self._release_type = ReleaseType(1 << pre_conv.get(lbl, 1))
        # All git versions are considered to have a dev part.  Contrary to python
        # versioning, a version with a dev number is greater than the same version
        # without one.
        self._has_dev_part = True
        self._dev_count = int(version_info["dev_n"] or 0)
        if self._dev_count > 0:
            self._release_type |= ReleaseType.DEV
        if self._release_type.value == 0:
            self._release_type = ReleaseType.FINAL
        self._is_inferred = version_info["inferred"] is not None
        self._is_dirty = version_info["dirty"] is not None

    @property
    def dirty(self) -> bool:
        return self._is_dirty

    @property
    def inferred(self) -> bool:
        return self._is_inferred

    def is_fallback(self) -> bool:
        return self._is_inferred and not self._valid_version

    def infer_last_tag(self) -> str:
        if self._valid_version:
            if self._is_inferred:
                # We can't infer a previous release from another inferred release
                return self._tag
            type_choices: Dict[int, str] = dict([(1, "a"), (2, "b"), (3, "rc")])
            if self.is_pre_release() and self._extra_tup > (0, 1, 0):
                type_idx = self._extra_tup[1]
                type_count = self._extra_tup[2]
                if type_count == 0:
                    type_idx -= 1
                else:
                    type_count -= 1
                pretype = type_choices.get(type_idx, "rc")
                return f"{self._release}.{pretype}{type_count}"
            else:
                parts = [int(ver) for ver in self._release.split(".")]
                new_ver: List[str] = []
                need_decrement = True
                for part in reversed(parts):
                    if part > 0 and need_decrement:
                        need_decrement = False
                        part -= 1
                    new_ver.insert(0, str(part))
                return "v" + ".".join(new_ver)
        return "v0.0.0"



# DBus Connection Management
#
# Copyright (C) 2022 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.
from __future__ import annotations
import os
import asyncio
import pathlib
import logging
import dbus_next
from dbus_next.aio import MessageBus, ProxyInterface
from dbus_next.constants import BusType

# Annotation imports
from typing import (
    TYPE_CHECKING,
    List,
    Optional,
    Any,
)

if TYPE_CHECKING:
    from ..confighelper import ConfigHelper

STAT_PATH = "/proc/self/stat"
DOC_URL = (
    "https://moonraker.readthedocs.io/en/latest/"
    "installation/#policykit-permissions"
)

class DbusManager:
    Variant = dbus_next.Variant
    DbusError = dbus_next.errors.DBusError
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.bus: Optional[MessageBus] = None
        self.polkit: Optional[ProxyInterface] = None
        self.warned: bool = False
        st_path = pathlib.Path(STAT_PATH)
        self.polkit_subject: List[Any] = []
        if not st_path.is_file():
            return
        proc_data = st_path.read_text()
        start_clk_ticks = int(proc_data.split()[21])
        self.polkit_subject = [
            "unix-process",
            {
                "pid": dbus_next.Variant("u", os.getpid()),
                "start-time": dbus_next.Variant("t", start_clk_ticks)
            }
        ]

    def is_connected(self) -> bool:
        return self.bus is not None and self.bus.connected

    async def component_init(self) -> None:
        try:
            self.bus = MessageBus(bus_type=BusType.SYSTEM)
            await self.bus.connect()
        except asyncio.CancelledError:
            raise
        except Exception:
            logging.info("Unable to Connect to D-Bus")
            return
        # Make sure that all required actions are register
        try:
            self.polkit = await self.get_interface(
                "org.freedesktop.PolicyKit1",
                "/org/freedesktop/PolicyKit1/Authority",
                "org.freedesktop.PolicyKit1.Authority")
        except asyncio.CancelledError:
            raise
        except Exception as e:
            if self.server.is_debug_enabled():
                logging.exception("Failed to get PolKit interface")
            else:
                logging.info(f"Failed to get PolKit interface: {e}")
            self.polkit = None

    async def check_permission(self,
                               action: str,
                               err_msg: str = ""
                               ) -> bool:
        if self.polkit is None:
            self.server.add_warning(
                "Unable to find DBus PolKit Interface, this suggests PolKit "
                "is not installed on your OS.",
                "dbus_polkit"
            )
            return False
        try:
            ret = await self.polkit.call_check_authorization(  # type: ignore
                self.polkit_subject, action, {}, 0, "")
        except asyncio.CancelledError:
            raise
        except Exception as e:
            self._check_warned()
            self.server.add_warning(
                f"Error checking authorization for action [{action}]: {e}. "
                "This suggests that a dependency is not installed or "
                f"up to date. {err_msg}.")
            return False
        if not ret[0]:
            self._check_warned()
            self.server.add_warning(
                "Moonraker not authorized for PolicyKit action: "
                f"[{action}], {err_msg}")
        return ret[0]

    def _check_warned(self):
        if not self.warned:
            self.server.add_warning(
                f"PolKit warnings detected. See {DOC_URL} for instructions "
                "on how to resolve.")
            self.warned = True

    async def get_interface(self,
                            bus_name: str,
                            bus_path: str,
                            interface_name: str
                            ) -> ProxyInterface:
        ret = await self.get_interfaces(bus_name, bus_path,
                                        [interface_name])
        return ret[0]

    async def get_interfaces(self,
                             bus_name: str,
                             bus_path: str,
                             interface_names: List[str]
                             ) -> List[ProxyInterface]:
        if self.bus is None:
            raise self.server.error("Bus not avaialable")
        interfaces: List[ProxyInterface] = []
        introspection = await self.bus.introspect(bus_name, bus_path)
        proxy_obj = self.bus.get_proxy_object(bus_name, bus_path,
                                              introspection)
        for ifname in interface_names:
            intf = proxy_obj.get_interface(ifname)
            interfaces.append(intf)
        return interfaces

    async def close(self):
        if self.bus is not None and self.bus.connected:
            self.bus.disconnect()
            await self.bus.wait_for_disconnect()


def load_component(config: ConfigHelper) -> DbusManager:
    return DbusManager(config)



# Git Deployment implementation
#
# Copyright (C) 2021  Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.

from __future__ import annotations
import asyncio
import os
import pathlib
import shutil
import re
import logging
from .app_deploy import AppDeploy
from .common import Channel
from ...utils.versions import GitVersion

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    Dict,
    List,
)
if TYPE_CHECKING:
    from ...confighelper import ConfigHelper
    from ..shell_command import ShellCommand
    from .update_manager import CommandHelper
    from ..http_client import HttpClient

class GitDeploy(AppDeploy):
    def __init__(self, config: ConfigHelper, cmd_helper: CommandHelper) -> None:
        super().__init__(config, cmd_helper, "Git Repo")
        self._configure_path(config)
        self._configure_virtualenv(config)
        self._configure_dependencies(config)
        self._configure_managed_services(config)
        self.origin: str = config.get('origin')
        self.moved_origin: Optional[str] = config.get('moved_origin', None)
        self.primary_branch = config.get("primary_branch", "master")
        pinned_commit = config.get("pinned_commit", None)
        if pinned_commit is not None:
            pinned_commit = pinned_commit.lower()
            # validate the hash length
            if len(pinned_commit) < 8:
                raise config.error(
                    f"[{config.get_name()}]: Value for option 'commit' must be "
                    "a minimum of 8 characters."
                )
        self.repo = GitRepo(
            cmd_helper, self.path, self.name, self.origin, self.moved_origin,
            self.primary_branch, self.channel, pinned_commit
        )

    async def initialize(self) -> Dict[str, Any]:
        storage = await super().initialize()
        await self.repo.restore_state(storage)
        self._is_valid = storage.get("is_valid", self.repo.is_valid())
        if not self.needs_refresh():
            self.repo.log_repo_info()
        return storage

    async def refresh(self) -> None:
        await self._update_repo_state(raise_exc=False)

    async def _update_repo_state(
        self, need_fetch: bool = True, raise_exc: bool = True
    ) -> None:
        self._is_valid = False
        try:
            await self.repo.refresh_repo_state(need_fetch=need_fetch)
        except Exception as e:
            if raise_exc or isinstance(e, asyncio.CancelledError):
                raise
        else:
            self._is_valid = self.repo.is_valid()
        finally:
            self.log_info(f"Channel: {self.channel}")
            if not self._is_valid:
                self.log_info("Repo validation check failed, updates disabled")
            else:
                self.log_info("Validity check for git repo passed")
            self._save_state()

    async def update(self) -> bool:
        await self.repo.wait_for_init()
        if not self._is_valid:
            raise self.log_exc("Update aborted, repo not valid", False)
        if self.repo.is_dirty():
            raise self.log_exc(
                "Update aborted, repo has been modified", False)
        if self.repo.is_current():
            # No need to update
            return False
        self.cmd_helper.notify_update_response(
            f"Updating Application {self.name}...")
        dep_info = await self._collect_dependency_info()
        await self._pull_repo()
        # Check Semantic Versions
        await self._update_dependencies(dep_info)
        # Refresh local repo state
        await self._update_repo_state(need_fetch=False)
        await self.restart_service()
        self.notify_status("Update Finished...", is_complete=True)
        return True

    async def recover(self,
                      hard: bool = False,
                      force_dep_update: bool = False
                      ) -> None:
        self.notify_status("Attempting Repo Recovery...")
        dep_info = await self._collect_dependency_info()
        if hard:
            await self.repo.clone()
            await self._update_repo_state()
        else:
            self.notify_status("Resetting Git Repo...")
            reset_ref = await self.repo.get_recovery_ref()
            if self.repo.is_dirty():
                # Try to restore modified files.  If the attempt fails we
                # can still try the reset
                try:
                    await self.repo.checkout("-- .")
                except self.server.error:
                    pass
            await self.repo.checkout(self.primary_branch)
            await self.repo.reset(reset_ref)
            await self._update_repo_state()
        self.repo.set_rollback_state(None)

        if self.repo.is_dirty() or not self._is_valid:
            raise self.server.error(
                "Recovery attempt failed, repo state not pristine", 500)
        await self._update_dependencies(dep_info, force=force_dep_update)
        await self.restart_service()
        self.notify_status("Reinstall Complete", is_complete=True)

    async def rollback(self) -> bool:
        dep_info = await self._collect_dependency_info()
        ret = await self.repo.rollback()
        if ret:
            await self._update_dependencies(dep_info)
            await self._update_repo_state(need_fetch=False)
            await self.restart_service()
            msg = "Rollback Complete"
        else:
            msg = "Rollback not performed"
        self.notify_status(msg, is_complete=True)
        return ret

    def get_update_status(self) -> Dict[str, Any]:
        status = super().get_update_status()
        status.update(self.repo.get_repo_status())
        return status

    def get_persistent_data(self) -> Dict[str, Any]:
        storage = super().get_persistent_data()
        storage.update(self.repo.get_persistent_data())
        return storage

    async def _pull_repo(self) -> None:
        self.notify_status("Updating Repo...")
        rb_state = self.repo.capture_state_for_rollback()
        try:
            await self.repo.fetch()
            if self.repo.is_detached():
                await self.repo.checkout()
            elif await self.repo.check_diverged():
                self.notify_status(
                    "Repo has diverged, attempting git reset"
                )
                await self.repo.reset()
            else:
                await self.repo.pull()
        except Exception as e:
            if self.repo.repo_corrupt:
                self._is_valid = False
                self._save_state()
                event_loop = self.server.get_event_loop()
                event_loop.delay_callback(
                    .2, self.cmd_helper.notify_update_refreshed
                )
            raise self.log_exc(str(e))
        else:
            self.repo.set_rollback_state(rb_state)


GIT_ASYNC_TIMEOUT = 300.
GIT_ENV_VARS = {
    'GIT_HTTP_LOW_SPEED_LIMIT': "1000",
    'GIT_HTTP_LOW_SPEED_TIME ': "20"
}
GIT_MAX_LOG_CNT = 100
GIT_LOG_FMT = (
    "\"sha:%H%x1Dauthor:%an%x1Ddate:%ct%x1Dsubject:%s%x1Dmessage:%b%x1E\""
)
GIT_REF_FMT = (
    "'%(if)%(*objecttype)%(then)%(*objecttype) %(*objectname)"
    "%(else)%(objecttype) %(objectname)%(end) %(refname)'"
)
SRC_EXTS = (".py", ".c", ".cpp")

class GitRepo:
    def __init__(
        self,
        cmd_helper: CommandHelper,
        src_path: pathlib.Path,
        alias: str,
        origin_url: str,
        moved_origin_url: Optional[str],
        primary_branch: str,
        channel: Channel,
        pinned_commit: Optional[str]
    ) -> None:
        self.server = cmd_helper.get_server()
        self.cmd_helper = cmd_helper
        self.alias = alias
        self.src_path = src_path
        git_dir = src_path.parent
        git_base = src_path.name
        self.backup_path = git_dir.joinpath(f".{git_base}_repo_backup")
        self.git_folder_path = src_path.joinpath(".git")
        self.origin_url = origin_url
        if not self.origin_url.endswith(".git"):
            self.origin_url += ".git"
        self.moved_origin_url = moved_origin_url
        self.primary_branch = primary_branch
        self.recovery_message = \
            f"""
            Manually restore via SSH with the following commands:
            sudo service {self.alias} stop
            cd {git_dir}
            rm -rf {git_base}
            git clone {self.origin_url}
            sudo service {self.alias} start
            """

        self.repo_warnings: List[str] = []
        self.repo_anomalies: List[str] = []
        self.init_evt: Optional[asyncio.Event] = None
        self.initialized: bool = False
        self.git_operation_lock = asyncio.Lock()
        self.fetch_timeout_handle: Optional[asyncio.Handle] = None
        self.fetch_input_recd: bool = False
        self.channel = channel
        self.pinned_commit = pinned_commit
        self.is_shallow = False

    async def restore_state(self, storage: Dict[str, Any]) -> None:
        self.valid_git_repo: bool = storage.get('repo_valid', False)
        self.git_owner: str = storage.get('git_owner', "?")
        self.git_repo_name: str = storage.get('git_repo_name', "?")
        self.git_remote: str = storage.get('git_remote', "?")
        self.git_branch: str = storage.get('git_branch', "?")
        if "full_version_string" in storage:
            self.current_version = GitVersion(storage["full_version_string"])
        else:
            self.current_version = GitVersion(storage.get('current_version', "?"))
        self.upstream_version = GitVersion(storage.get('upstream_version', "?"))
        self.current_commit: str = storage.get('current_commit', "?")
        self.upstream_commit: str = storage.get('upstream_commit', "?")
        self.upstream_url: str = storage.get('upstream_url', "?")
        self.recovery_url: str = storage.get(
            'recovery_url',
            self.upstream_url if self.git_remote == "origin" else "?"
        )
        self.branches: List[str] = storage.get('branches', [])
        self.head_detached: bool = storage.get('head_detached', False)
        self.git_messages: List[str] = storage.get('git_messages', [])
        self.commits_behind: List[Dict[str, Any]] = storage.get('commits_behind', [])
        self.commits_behind_count: int = storage.get('cbh_count', 0)
        self.diverged: bool = storage.get("diverged", False)
        self.repo_corrupt: bool = storage.get('corrupt', False)
        self.modified_files: List[str] = storage.get("modified_files", [])
        self.untracked_files: List[str] = storage.get("untracked_files", [])
        def_rbs = self.capture_state_for_rollback()
        self.rollback_commit: str = storage.get('rollback_commit', self.current_commit)
        self.rollback_branch: str = storage.get('rollback_branch', def_rbs["branch"])
        rbv = storage.get('rollback_version', self.current_version)
        self.rollback_version = GitVersion(str(rbv))
        self.pinned_commit_valid: bool = storage.get('pinned_commit_valid', True)
        if not await self._detect_git_dir():
            self.valid_git_repo = False
        self._check_warnings()

    def get_persistent_data(self) -> Dict[str, Any]:
        return {
            'repo_valid': self.valid_git_repo,
            'git_owner': self.git_owner,
            'git_repo_name': self.git_repo_name,
            'git_remote': self.git_remote,
            'git_branch': self.git_branch,
            'current_version': self.current_version.full_version,
            'upstream_version': self.upstream_version.full_version,
            'current_commit': self.current_commit,
            'upstream_commit': self.upstream_commit,
            'rollback_commit': self.rollback_commit,
            'rollback_branch': self.rollback_branch,
            'rollback_version': self.rollback_version.full_version,
            'upstream_url': self.upstream_url,
            'recovery_url': self.recovery_url,
            'branches': self.branches,
            'head_detached': self.head_detached,
            'git_messages': self.git_messages,
            'commits_behind': self.commits_behind,
            'cbh_count': self.commits_behind_count,
            'diverged': self.diverged,
            'corrupt': self.repo_corrupt,
            'modified_files': self.modified_files,
            'untracked_files': self.untracked_files,
            'pinned_commit_valid': self.pinned_commit_valid
        }

    async def refresh_repo_state(self, need_fetch: bool = True) -> None:
        if self.init_evt is not None:
            # No need to initialize multiple requests
            await self.init_evt.wait()
            if self.initialized:
                return
        self.initialized = False
        self.pinned_commit_valid = True
        self.init_evt = asyncio.Event()
        self.git_messages.clear()
        try:
            await self._check_repo_status()
            self._verify_repo()
            await self._find_current_branch()

            # Fetch the upstream url.  If the repo has been moved,
            # set the new url
            self.upstream_url = await self.remote(f"get-url {self.git_remote}", True)
            if await self._check_moved_origin():
                need_fetch = True
            if self.git_remote == "origin":
                self.recovery_url = self.upstream_url
            else:
                remote_list = (await self.remote()).splitlines()
                logging.debug(
                    f"Git Repo {self.alias}: Detected Remotes - {remote_list}"
                )
                if "origin" in remote_list:
                    self.recovery_url = await self.remote("get-url origin")
                else:
                    logging.info(
                        f"Git Repo {self.alias}: Unable to detect recovery URL, "
                        "Hard Recovery not available"
                    )
                    self.recovery_url = "?"
            if need_fetch:
                await self.fetch()
            self.diverged = await self.check_diverged()

            # Parse GitHub Owner from URL
            owner_match = re.match(r"https?://[^/]+/([^/]+)", self.upstream_url)
            self.git_owner = "?"
            if owner_match is not None:
                self.git_owner = owner_match.group(1)

            # Parse GitHub Repository Name from URL
            repo_match = re.match(r".*\/([^\.]*).*", self.upstream_url)
            self.git_repo_name = "?"
            if repo_match is not None:
                self.git_repo_name = repo_match.group(1)
            self.current_commit = await self.rev_parse("HEAD")
            git_desc = await self.describe("--always --tags --long --dirty --abbrev=8")
            cur_ver = GitVersion(git_desc.strip())
            upstream_ver = await self._get_upstream_version()
            await self._set_versions(cur_ver, upstream_ver)

            # Get Commits Behind
            self.commits_behind = []
            if self.commits_behind_count > 0:
                cbh = await self.get_commits_behind()
                tagged_commits = await self.get_tagged_commits()
                debug_msg = '\n'.join([f"{k}: {v}" for k, v in tagged_commits.items()])
                logging.debug(f"Git Repo {self.alias}: Tagged Commits\n{debug_msg}")
                for i, commit in enumerate(cbh):
                    tag = tagged_commits.get(commit['sha'], None)
                    if i < 30 or tag is not None:
                        commit['tag'] = tag
                        self.commits_behind.append(commit)
            self._check_warnings()
        except Exception:
            logging.exception(f"Git Repo {self.alias}: Initialization failure")
            self._check_warnings()
            raise
        else:
            self.initialized = True
            # If no exception was raised assume the repo is not corrupt
            self.repo_corrupt = False
            if self.rollback_commit == "?" or self.rollback_branch == "?":
                # Reset Rollback State
                self.set_rollback_state(None)
            self.log_repo_info()
        finally:
            self.init_evt.set()
            self.init_evt = None

    async def _check_repo_status(self) -> bool:
        async with self.git_operation_lock:
            self.valid_git_repo = False
            if not await self._detect_git_dir():
                logging.info(
                    f"Git Repo {self.alias}: path '{self.src_path}'"
                    " is not a valid git repo")
                return False
            await self._wait_for_lock_release()
            attempts = 3
            resp: Optional[str] = None
            while attempts:
                self.git_messages.clear()
                try:
                    cmd = "status --porcelain -b"
                    resp = await self._run_git_cmd(
                        cmd, attempts=1, corrupt_hdr="fatal:"
                    )
                except Exception:
                    attempts -= 1
                    resp = None
                    # Attempt to recover from "loose object" error
                    if attempts and self.repo_corrupt:
                        if not await self._repair_loose_objects():
                            # Since we are unable to recover, immediately
                            # return
                            return False
                else:
                    break
            if resp is None:
                return False
            self.modified_files.clear()
            self.untracked_files.clear()
            for line in resp.splitlines():
                parts = line.strip().split(maxsplit=1)
                if len(parts) != 2:
                    continue
                prefix, fname = [p.strip() for p in parts]
                if prefix == "M":
                    # modified file
                    self.modified_files.append(fname)
                elif prefix == "??":
                    # untracked file
                    ext = pathlib.Path(fname).suffix
                    if ext in SRC_EXTS:
                        self.untracked_files.append(fname)
            self.valid_git_repo = True
        return True

    async def _detect_git_dir(self) -> bool:
        if self.git_folder_path.is_file():
            # Submodules have a file that contain the path to
            # the .git folder
            eventloop = self.server.get_event_loop()
            data = await eventloop.run_in_thread(self.git_folder_path.read_text)
            ident, _, gitdir = data.partition(":")
            if ident.strip() != "gitdir" or not gitdir.strip():
                return False
            self.git_folder_path = pathlib.Path(gitdir).expanduser().resolve()
        if self.git_folder_path.is_dir():
            self.is_shallow = self.git_folder_path.joinpath("shallow").is_file()
            return True
        return False

    async def _find_current_branch(self) -> None:
        # Populate list of current branches
        blist = await self.list_branches()
        current_branch = ""
        self.branches = []
        for branch in blist:
            branch = branch.strip()
            if not branch:
                continue
            if branch[0] == "*":
                branch = branch[2:].strip()
                current_branch = branch
            if branch[0] == "(":
                continue
            self.branches.append(branch)
        if current_branch.startswith("(HEAD detached"):
            self.head_detached = True
            ref_name = current_branch.split()[-1][:-1]
            remote_list = (await self.remote()).splitlines()
            for remote in remote_list:
                remote = remote.strip()
                if not remote:
                    continue
                if ref_name.startswith(remote):
                    self.git_branch = ref_name[len(remote)+1:]
                    self.git_remote = remote
                    break
            else:
                if self.git_remote == "?":
                    msg = "Resolve by manually checking out a branch via SSH."
                else:
                    prev = f"{self.git_remote}/{self.git_branch}"
                    msg = f"Defaulting to previously tracked {prev}."
                logging.info(f"Git Repo {self.alias}: {current_branch} {msg}")
        else:
            self.head_detached = False
            self.git_branch = current_branch
            rkey = f"branch.{self.git_branch}.remote"
            self.git_remote = (await self.config_get(rkey)) or "?"

    async def _check_moved_origin(self) -> bool:
        detected_origin = self.upstream_url.lower().strip()
        if not detected_origin.endswith(".git"):
            detected_origin += ".git"
        if (
            self.server.is_debug_enabled() or
            not detected_origin.startswith("http") or
            detected_origin == self.origin_url.lower()
        ):
            # Skip the moved origin check if:
            #  Repo Debug is enabled
            #  The detected origin url is not http(s)
            #  The detected origin matches the expected origin url
            return False
        moved = False
        client: HttpClient = self.server.lookup_component("http_client")
        check_url = detected_origin[:-4]
        logging.info(
            f"Git repo {self.alias}: Performing moved origin check - "
            f"{check_url}"
        )
        resp = await client.get(check_url, enable_cache=False)
        if not resp.has_error():
            final_url = resp.final_url.lower()
            if not final_url.endswith(".git"):
                final_url += ".git"
            logging.debug(f"Git repo {self.alias}: Resolved url - {final_url}")
            if final_url == self.origin_url.lower():
                logging.info(
                    f"Git Repo {self.alias}: Moved Repo Detected, Moving "
                    f"from {self.upstream_url} to {self.origin_url}")
                moved = True
                await self.remote(
                    f"set-url {self.git_remote} {self.origin_url}", True
                )
                self.upstream_url = self.origin_url
                if self.moved_origin_url is not None:
                    moved_origin = self.moved_origin_url.lower().strip()
                    if not moved_origin.endswith(".git"):
                        moved_origin += ".git"
                    if moved_origin != detected_origin:
                        self.server.add_warning(
                            f"Git Repo {self.alias}: Origin URL does not "
                            "not match configured 'moved_origin'option. "
                            f"Expected: {detected_origin}"
                        )
        else:
            logging.debug(f"Move Request Failed: {resp.error}")
        return moved

    async def _get_upstream_version(self) -> GitVersion:
        self.commits_behind_count = 0
        if self.pinned_commit is not None:
            self.upstream_commit = self.current_commit
            if not self.current_commit.lower().startswith(self.pinned_commit):
                if not await self.check_commit_exists(self.pinned_commit):
                    self.pinned_commit_valid = False
                elif await self.is_ancestor(self.current_commit, self.pinned_commit):
                    self.upstream_commit = self.pinned_commit
            upstream_ver_str = await self.describe(
                f"{self.upstream_commit} --always --tags --long --abbrev=8",
            )
        elif self.channel == Channel.DEV:
            self.upstream_commit = await self.rev_parse(
                f"{self.git_remote}/{self.git_branch}"
            )
            upstream_ver_str = await self.describe(
                f"{self.git_remote}/{self.git_branch} --always --tags --long --abbrev=8"
            )
        else:
            tagged_commits = await self.get_tagged_commits()
            upstream_commit = upstream_ver_str = "?"
            for sha, tag in tagged_commits.items():
                ver = GitVersion(tag)
                if not ver.is_valid_version():
                    continue
                if (
                    (self.channel == Channel.STABLE and ver.is_final_release()) or
                    (self.channel == Channel.BETA and not ver.is_alpha_release())
                ):
                    upstream_commit = sha
                    upstream_ver_str = tag
                    break
            self.upstream_commit = upstream_commit
        if self.upstream_commit != "?":
            rl_args = f"HEAD..{self.upstream_commit} --count"
            self.commits_behind_count = int(await self.rev_list(rl_args))
        return GitVersion(upstream_ver_str)

    async def _set_versions(
        self, current_version: GitVersion, upstream_version: GitVersion
    ) -> None:
        if not current_version.is_valid_version():
            log_msg = (
                f"Git repo {self.alias}: Failed to detect current version, got "
                f"'{current_version}'. "
            )
            tag = upstream_version.infer_last_tag()
            count = await self.rev_list("HEAD --count")
            sha_part = ""
            if current_version.is_fallback():
                sha_part = f"-g{current_version}"
            elif self.current_commit not in ("?", ""):
                sha_part = f"-g{self.current_commit[:8]}"
            current_version = GitVersion(f"{tag}-{count}{sha_part}-inferred")
            log_msg += f"Falling back to inferred version: {current_version}"
            logging.info(log_msg)
        if self.channel == Channel.DEV:
            if not upstream_version.is_valid_version():
                log_msg = (
                    f"Git repo {self.alias}: Failed to detect upstream version, got "
                    f"'{upstream_version}'. "
                )
                tag = current_version.tag
                if current_version.inferred:
                    count = await self.rev_list(f"{self.upstream_commit} --count")
                else:
                    log_msg += "\nRemote has diverged, approximating dev count. "
                    count = str(self.commits_behind_count + current_version.dev_count)
                upstream_version = GitVersion(f"{tag}-{count}-inferred")
                log_msg += f"Falling back to inferred version: {upstream_version}"
                logging.info(log_msg)
        else:
            if not upstream_version.is_valid_version():
                self.upstream_commit = self.current_commit
                upstream_version = current_version
            elif upstream_version <= current_version:
                self.upstream_commit = self.current_commit
        self.current_version = current_version
        self.upstream_version = upstream_version

    async def wait_for_init(self) -> None:
        if self.init_evt is not None:
            await self.init_evt.wait()
            if not self.initialized:
                raise self.server.error(
                    f"Git Repo {self.alias}: Initialization failure")

    async def is_ancestor(
        self, ancestor_ref: str, descendent_ref: str, attempts: int = 3
    ) -> bool:
        self._verify_repo()
        cmd = f"merge-base --is-ancestor {ancestor_ref} {descendent_ref}"
        async with self.git_operation_lock:
            for _ in range(attempts):
                try:
                    await self._run_git_cmd(cmd, attempts=1, corrupt_hdr="error: ")
                except self.cmd_helper.get_shell_command().error as err:
                    if err.return_code == 1:
                        return False
                    if self.repo_corrupt:
                        raise
                else:
                    break
                await asyncio.sleep(.2)
            return True

    async def check_diverged(self) -> bool:
        self._verify_repo(check_remote=True)
        if self.head_detached:
            return False
        descendent = f"{self.git_remote}/{self.git_branch}"
        return not (await self.is_ancestor("HEAD", descendent))

    def log_repo_info(self) -> None:
        warnings = self._generate_warn_msg()
        if warnings:
            warnings = "\nRepo Warnings:\n" + warnings
        logging.info(
            f"Git Repo {self.alias} Detected:\n"
            f"Owner: {self.git_owner}\n"
            f"Repository Name: {self.git_repo_name}\n"
            f"Path: {self.src_path}\n"
            f"Remote: {self.git_remote}\n"
            f"Branch: {self.git_branch}\n"
            f"Remote URL: {self.upstream_url}\n"
            f"Recovery URL: {self.recovery_url}\n"
            f"Current Commit SHA: {self.current_commit}\n"
            f"Upstream Commit SHA: {self.upstream_commit}\n"
            f"Current Version: {self.current_version}\n"
            f"Upstream Version: {self.upstream_version}\n"
            f"Rollback Commit: {self.rollback_commit}\n"
            f"Rollback Branch: {self.rollback_branch}\n"
            f"Rollback Version: {self.rollback_version}\n"
            f"Is Dirty: {self.current_version.dirty}\n"
            f"Is Detached: {self.head_detached}\n"
            f"Is Shallow: {self.is_shallow}\n"
            f"Commits Behind Count: {self.commits_behind_count}\n"
            f"Diverged: {self.diverged}\n"
            f"Pinned Commit: {self.pinned_commit}"
            f"{warnings}"
        )

    def _check_warnings(self) -> None:
        self.repo_warnings.clear()
        self.repo_anomalies.clear()
        if self.pinned_commit is not None and not self.pinned_commit_valid:
            self.repo_anomalies.append(
                f"Pinned Commit {self.pinned_commit} does not exist"
            )
        if self.repo_corrupt:
            self.repo_warnings.append("Repo is corrupt")
        if self.git_branch == "?":
            self.repo_warnings.append("Failed to detect git branch")
        elif self.git_remote == "?":
            self.repo_warnings.append(
                f"Failed to detect tracking remote for branch {self.git_branch}"
            )
        if self.upstream_url == "?":
            self.repo_warnings.append("Failed to detect repo url")
            return
        upstream_url = self.upstream_url.lower()
        if upstream_url[-4:] != ".git":
            upstream_url += ".git"
        if upstream_url != self.origin_url.lower():
            self.repo_anomalies.append(f"Unofficial remote url: {self.upstream_url}")
        if self.git_branch != self.primary_branch or self.git_remote != "origin":
            self.repo_anomalies.append(
                "Repo not on offical remote/branch, expected: "
                f"origin/{self.primary_branch}, detected: "
                f"{self.git_remote}/{self.git_branch}")
        if self.untracked_files:
            self.repo_anomalies.append(
                f"Repo has untracked source files: {self.untracked_files}"
            )
        if self.diverged:
            self.repo_anomalies.append("Repo has diverged from remote")
        if self.head_detached:
            msg = "Detached HEAD detected"
            if self.server.is_debug_enabled():
                self.repo_anomalies.append(msg)
            else:
                self.repo_warnings.append(msg)
        if self.is_dirty():
            self.repo_warnings.append(
                "Repo is dirty.  Detected the following modifed files: "
                f"{self.modified_files}"
            )
        self._generate_warn_msg()

    def _generate_warn_msg(self) -> str:
        ro_msg = f"Git Repo {self.alias}: No warnings detected"
        warn_msg = ""
        if self.repo_warnings or self.repo_anomalies:
            ro_msg = f"Git Repo {self.alias}: Warnings detected:\n"
            warn_msg = "\n".join(
                [f"  {warn}" for warn in self.repo_warnings + self.repo_anomalies]
            )
            ro_msg += warn_msg
        self.server.add_log_rollover_item(f"umgr_{self.alias}_warn", ro_msg, log=False)
        return warn_msg

    def _verify_repo(self, check_remote: bool = False) -> None:
        if not self.valid_git_repo:
            raise self.server.error(
                f"Git Repo {self.alias}: repo not initialized")
        if check_remote:
            if self.git_remote == "?":
                raise self.server.error(
                    f"Git Repo {self.alias}: No valid git remote detected")

    async def reset(self, ref: Optional[str] = None) -> None:
        async with self.git_operation_lock:
            if ref is None:
                if self.channel != Channel.DEV or self.pinned_commit is not None:
                    ref = self.upstream_commit
                else:
                    if self.git_remote == "?" or self.git_branch == "?":
                        raise self.server.error("Cannot reset, unknown remote/branch")
                    ref = f"{self.git_remote}/{self.git_branch}"
            await self._run_git_cmd(f"reset --hard {ref}", attempts=2)
            self.repo_corrupt = False

    async def fetch(self) -> None:
        self._verify_repo(check_remote=True)
        async with self.git_operation_lock:
            await self._run_git_cmd_async(
                f"fetch {self.git_remote} --prune --progress")

    async def clean(self) -> None:
        self._verify_repo()
        async with self.git_operation_lock:
            await self._run_git_cmd("clean -d -f", attempts=2)

    async def pull(self) -> None:
        self._verify_repo()
        if self.head_detached:
            raise self.server.error(
                f"Git Repo {self.alias}: Cannot perform pull on a "
                "detached HEAD")
        cmd = "pull --progress"
        if self.server.is_debug_enabled():
            cmd = f"{cmd} --rebase"
        if self.channel != Channel.DEV or self.pinned_commit is not None:
            cmd = f"{cmd} {self.git_remote} {self.upstream_commit}"
        async with self.git_operation_lock:
            await self._run_git_cmd_async(cmd)

    async def list_branches(self) -> List[str]:
        self._verify_repo()
        async with self.git_operation_lock:
            resp = await self._run_git_cmd("branch --list --no-color")
            return resp.strip().split("\n")

    async def check_commit_exists(self, commit: str) -> bool:
        self._verify_repo()
        async with self.git_operation_lock:
            shell_cmd = self.cmd_helper.get_shell_command()
            try:
                await self._run_git_cmd(
                    f"cat-file -e {commit}^{{commit}}", attempts=1
                )
            except shell_cmd.error:
                return False
            return True

    async def remote(self, command: str = "", validate: bool = False) -> str:
        self._verify_repo(check_remote=validate)
        async with self.git_operation_lock:
            resp = await self._run_git_cmd(
                f"remote {command}")
            return resp.strip()

    async def describe(self, args: str = "") -> str:
        self._verify_repo()
        async with self.git_operation_lock:
            resp = await self._run_git_cmd(f"describe {args}".strip())
            return resp.strip()

    async def rev_parse(self, args: str = "") -> str:
        self._verify_repo()
        async with self.git_operation_lock:
            resp = await self._run_git_cmd(f"rev-parse {args}".strip())
            return resp.strip()

    async def rev_list(self, args: str = "") -> str:
        self._verify_repo()
        async with self.git_operation_lock:
            resp = await self._run_git_cmd(f"rev-list {args}".strip())
            return resp.strip()

    async def config_get(
        self,
        key: str,
        pattern: str = "",
        get_all: bool = False,
        local_only: bool = False
    ) -> Optional[str]:
        local = "--local " if local_only else ""
        cmd = f"{local}--get-all" if get_all else f"{local}--get"
        args = f"{cmd} {key} '{pattern}'" if pattern else f"{cmd} {key}"
        try:
            return await self.config_cmd(args)
        except self.cmd_helper.get_shell_command().error as e:
            if e.return_code == 1:
                return None
            raise

    async def config_set(self, key: str, value: str) -> None:
        await self.config_cmd(f"{key} '{value}'")

    async def config_add(self, key: str, value: str) -> None:
        await self.config_cmd(f"--add {key} '{value}'")

    async def config_unset(
        self, key: str, pattern: str = "", unset_all: bool = False
    ) -> None:
        cmd = "--unset-all" if unset_all else "--unset"
        args = f"{cmd} {key} '{pattern}'" if pattern else f"{cmd} {key}"
        await self.config_cmd(args)

    async def config_cmd(self, args: str) -> str:
        self._verify_repo()
        verbose = self.server.is_verbose_enabled()
        async with self.git_operation_lock:
            for attempt in range(3):
                try:
                    return await self._run_git_cmd(
                        f"config {args}", attempts=1, log_complete=verbose
                    )
                except self.cmd_helper.get_shell_command().error as e:
                    if 1 <= (e.return_code or 10) <= 6 or attempt == 2:
                        raise
            raise self.server.error("Failed to run git-config")


    async def checkout(self, branch: Optional[str] = None) -> None:
        self._verify_repo()
        reset_commit: Optional[str] = None
        async with self.git_operation_lock:
            if branch is None:
                # No branch is specifed so we are checking out detached
                if self.channel != Channel.DEV or self.pinned_commit is not None:
                    reset_commit = self.upstream_commit
                branch = f"{self.git_remote}/{self.git_branch}"
            await self._run_git_cmd(f"checkout -q {branch}")
        if reset_commit is not None:
            await self.reset(reset_commit)

    async def run_fsck(self) -> None:
        async with self.git_operation_lock:
            await self._run_git_cmd("fsck --full", timeout=300., attempts=1)

    async def clone(self) -> None:
        if self.is_submodule_or_worktree():
            raise self.server.error(
                f"Cannot clone git repo {self.alias}, it is a {self.get_repo_type()} "
                "of another git repo."
            )
        async with self.git_operation_lock:
            if self.recovery_url == "?":
                raise self.server.error(
                    "Recovery url has not been detected, clone aborted"
                )
            self.cmd_helper.notify_update_response(
                f"Git Repo {self.alias}: Starting Clone Recovery...")
            event_loop = self.server.get_event_loop()
            if self.backup_path.exists():
                await event_loop.run_in_thread(shutil.rmtree, self.backup_path)
            await self._check_lock_file_exists(remove=True)
            cmd = (
                f"clone --branch {self.primary_branch} --filter=blob:none "
                f"{self.recovery_url} {self.backup_path}"
            )
            try:
                await self._run_git_cmd_async(cmd, 1, False, False)
            except Exception as e:
                self.cmd_helper.notify_update_response(
                    f"Git Repo {self.alias}: Git Clone Failed")
                raise self.server.error("Git Clone Error") from e
            if self.src_path.exists():
                await event_loop.run_in_thread(shutil.rmtree, self.src_path)
            await event_loop.run_in_thread(
                shutil.move, str(self.backup_path), str(self.src_path))
            self.repo_corrupt = False
            self.valid_git_repo = True
            self.cmd_helper.notify_update_response(
                f"Git Repo {self.alias}: Git Clone Complete")
        reset_commit = await self.get_recovery_ref("HEAD")
        if reset_commit != "HEAD":
            self.cmd_helper.notify_update_response(
                f"Git Repo {self.alias}: Moving HEAD to previous "
                f"commit {self.current_commit}"
            )
            await self.reset(reset_commit)

    async def rollback(self) -> bool:
        if self.rollback_commit == "?" or self.rollback_branch == "?":
            raise self.server.error("Incomplete rollback data stored, cannot rollback")
        if self.rollback_branch != self.git_branch:
            await self.checkout(self.rollback_branch)
        elif self.rollback_commit == self.current_commit:
            return False
        await self.reset(self.rollback_commit)
        return True

    def capture_state_for_rollback(self) -> Dict[str, Any]:
        branch = self.git_branch
        if self.head_detached:
            valid = "?" not in (self.git_remote, self.git_branch)
            branch = f"{self.git_remote}/{self.git_branch}" if valid else "?"
        return {
            "commit": self.current_commit,
            "branch": branch,
            "version": self.current_version
        }

    def set_rollback_state(self, rb_state: Optional[Dict[str, Any]]) -> None:
        if rb_state is None:
            rb_state = self.capture_state_for_rollback()
        self.rollback_commit = rb_state["commit"]
        self.rollback_branch = rb_state["branch"]
        self.rollback_version = rb_state["version"]

    async def get_commits_behind(self) -> List[Dict[str, Any]]:
        self._verify_repo()
        if self.is_current():
            return []
        async with self.git_operation_lock:
            if self.channel != Channel.DEV or self.pinned_commit is not None:
                ref = self.upstream_commit
            else:
                ref = f"{self.git_remote}/{self.git_branch}"
            resp = await self._run_git_cmd(
                f"log {self.current_commit}..{ref} "
                f"--format={GIT_LOG_FMT} --max-count={GIT_MAX_LOG_CNT}")
            commits_behind: List[Dict[str, Any]] = []
            for log_entry in resp.split('\x1E'):
                log_entry = log_entry.strip()
                if not log_entry:
                    continue
                log_items = [li.strip() for li in log_entry.split('\x1D')
                             if li.strip()]
                cbh = [li.split(':', 1) for li in log_items]
                commits_behind.append(dict(cbh))  # type: ignore
            return commits_behind

    async def get_tagged_commits(self, count: int = 100) -> Dict[str, str]:
        self._verify_repo(check_remote=True)
        tip = f"{self.git_remote}/{self.git_branch}"
        cnt_arg = f"--count={count} " if count > 0 else ""
        async with self.git_operation_lock:
            resp = await self._run_git_cmd(
                f"for-each-ref {cnt_arg}--sort='-creatordate' --contains=HEAD "
                f"--merged={tip} --format={GIT_REF_FMT} 'refs/tags'"
            )
            tagged_commits: Dict[str, str] = {}
            for line in resp.split('\n'):
                parts = line.strip().split()
                if len(parts) != 3 or parts[0] != "commit":
                    continue
                sha, ref = parts[1:]
                tag = ref.split('/')[-1]
                tagged_commits[sha] = tag
            # Return tagged commits as SHA keys mapped to tag values
            return tagged_commits

    def get_repo_status(self) -> Dict[str, Any]:
        no_untrk_src = len(self.untracked_files) == 0
        return {
            'detected_type': "git_repo",
            'remote_alias': self.git_remote,
            'branch': self.git_branch,
            'owner': self.git_owner,
            'repo_name': self.git_repo_name,
            'remote_url': self.upstream_url,
            'recovery_url': self.recovery_url,
            'version': self.current_version.short_version,
            'remote_version': self.upstream_version.short_version,
            'rollback_version': self.rollback_version.short_version,
            'current_hash': self.current_commit,
            'remote_hash': self.upstream_commit,
            'is_dirty': self.current_version.dirty,
            'detached': self.head_detached,
            'commits_behind': self.commits_behind,
            'commits_behind_count': self.commits_behind_count,
            'git_messages': self.git_messages,
            'full_version_string': self.current_version.full_version,
            'pristine': no_untrk_src and not self.current_version.dirty,
            'corrupt': self.repo_corrupt,
            'warnings': self.repo_warnings,
            'anomalies': self.repo_anomalies
        }

    def get_version(self, upstream: bool = False) -> GitVersion:
        return self.upstream_version if upstream else self.current_version

    def is_detached(self) -> bool:
        return self.head_detached

    def is_dirty(self) -> bool:
        return self.current_version.dirty

    def is_current(self) -> bool:
        return self.current_commit == self.upstream_commit

    def is_submodule_or_worktree(self):
        return (
            self.src_path.joinpath(".git").is_file() and
            self.git_folder_path.parent.name in ("modules", "worktrees")
        )

    def is_valid(self) -> bool:
        return (
            not self.is_damaged() and
            not self.has_recoverable_errors()
        )

    def is_damaged(self) -> bool:
        # A damaged repo requires a clone to recover
        return not self.valid_git_repo or self.repo_corrupt

    def has_recoverable_errors(self) -> bool:
        # These errors should be recoverable using a git reset
        detached_err = False if self.server.is_debug_enabled() else self.head_detached
        return (
            self.diverged or
            self.is_dirty() or
            detached_err
        )

    def get_repo_type(self) -> str:
        type_name = self.git_folder_path.parent.name
        if type_name == "modules":
            return "submodule"
        elif type_name == "worktrees":
            return "worktree"
        return "repo"

    async def get_recovery_ref(self, upstream_ref: Optional[str] = None) -> str:
        """ Fetch the best reference for a 'reset' recovery attempt

        Returns the ref to reset to for "soft" recovery requests.  The
        preference is to reset to the current commit, however that is
        only possible if the commit is known and if it is an ancestor of
        the primary branch.
        """
        if upstream_ref is None:
            remote = await self.config_get(f"branch.{self.primary_branch}.remote")
            if remote is None:
                raise self.server.error(
                    f"Failed to find remote for primary branch '{self.primary_branch}'"
                )
            upstream_ref = f"{remote}/{self.primary_branch}"
        reset_commits: List[str] = []
        if self.pinned_commit is not None:
            reset_commits.append(self.pinned_commit)
        if self.current_commit != "?":
            reset_commits.append(self.current_commit)
        for commit in reset_commits:
            try:
                is_ancs = await self.is_ancestor(commit, upstream_ref, attempts=1)
            except self.server.error:
                is_ancs = False
            if is_ancs:
                return commit
        return upstream_ref

    async def _check_lock_file_exists(self, remove: bool = False) -> bool:
        lock_path = self.git_folder_path.joinpath("index.lock")
        if lock_path.is_file():
            if remove:
                logging.info(f"Git Repo {self.alias}: Git lock file found "
                             "after git process exited, removing")
                try:
                    event_loop = self.server.get_event_loop()
                    await event_loop.run_in_thread(os.remove, lock_path)
                except Exception:
                    pass
            return True
        return False

    async def _wait_for_lock_release(self, timeout: int = 60) -> None:
        while timeout:
            if await self._check_lock_file_exists():
                if not timeout % 10:
                    logging.info(f"Git Repo {self.alias}: Git lock file "
                                 f"exists, {timeout} seconds remaining "
                                 "before removal.")
                await asyncio.sleep(1.)
                timeout -= 1
            else:
                return
        await self._check_lock_file_exists(remove=True)

    async def _repair_loose_objects(self, notify: bool = False) -> bool:
        if notify:
            self.cmd_helper.notify_update_response(
                "Attempting to repair loose objects..."
            )
        try:
            shell_cmd = self.cmd_helper.get_shell_command()
            await shell_cmd.exec_cmd(
                "find .git/objects/ -type f -empty | xargs rm",
                timeout=10., attempts=1, cwd=str(self.src_path))
            await self._run_git_cmd_async(
                "fetch --all -p", attempts=1, fix_loose=False)
            await self._run_git_cmd("fsck --full", timeout=300., attempts=1)
        except Exception:
            msg = (
                "Attempt to repair loose objects failed, "
                "hard recovery is required"
            )
            logging.exception(msg)
            if notify:
                self.cmd_helper.notify_update_response(msg)
            return False
        if notify:
            self.cmd_helper.notify_update_response("Loose objects repaired")
        self.repo_corrupt = False
        return True

    async def _run_git_cmd_async(self,
                                 cmd: str,
                                 attempts: int = 5,
                                 need_git_path: bool = True,
                                 fix_loose: bool = True
                                 ) -> None:
        # Fetch and pull require special handling.  If the request
        # gets delayed we do not want to terminate it while the command
        # is processing.
        await self._wait_for_lock_release()
        event_loop = self.server.get_event_loop()
        env = os.environ.copy()
        env.update(GIT_ENV_VARS)
        if need_git_path:
            git_cmd = f"git -C {self.src_path} {cmd}"
        else:
            git_cmd = f"git {cmd}"
        shell_cmd = self.cmd_helper.get_shell_command()
        scmd = shell_cmd.build_shell_command(
            git_cmd, callback=self._handle_process_output,
            env=env)
        while attempts:
            self.git_messages.clear()
            self.fetch_input_recd = False
            self.fetch_timeout_handle = event_loop.delay_callback(
                GIT_ASYNC_TIMEOUT, self._check_process_active,
                scmd, cmd)
            try:
                await scmd.run(timeout=0)
            except Exception:
                pass
            if self.fetch_timeout_handle is not None:
                self.fetch_timeout_handle.cancel()
            ret = scmd.get_return_code()
            if ret == 0:
                self.git_messages.clear()
                return
            elif self.repo_corrupt and fix_loose:
                if await self._repair_loose_objects(notify=True):
                    # Only attempt to repair loose objects once. Re-run
                    # the command once.
                    fix_loose = False
                    attempts = 2
                else:
                    # since the attept to repair failed, bypass attempts
                    # and immediately raise an exception
                    raise self.server.error(
                        "Unable to repair loose objects, use hard recovery"
                    )
            attempts -= 1
            await asyncio.sleep(.5)
            await self._check_lock_file_exists(remove=True)
        raise self.server.error(f"Git Command '{cmd}' failed")

    def _handle_process_output(self, output: bytes) -> None:
        self.fetch_input_recd = True
        out = output.decode().strip()
        if out:
            if out.startswith("fatal: ") and "corrupt" in out:
                self.repo_corrupt = True
            self.git_messages.append(out)
            self.cmd_helper.notify_update_response(out)
            logging.debug(
                f"Git Repo {self.alias}: {out}")

    async def _check_process_active(
        self, scmd: ShellCommand, cmd_name: str
    ) -> None:
        ret = scmd.get_return_code()
        if ret is not None:
            logging.debug(f"Git Repo {self.alias}: {cmd_name} returned")
            return
        if self.fetch_input_recd:
            # Received some input, reschedule timeout
            logging.debug(
                f"Git Repo {self.alias}: {cmd_name} active, rescheduling")
            event_loop = self.server.get_event_loop()
            self.fetch_input_recd = False
            self.fetch_timeout_handle = event_loop.delay_callback(
                GIT_ASYNC_TIMEOUT, self._check_process_active,
                scmd, cmd_name)
        else:
            # Request has timed out with no input, terminate it
            logging.debug(f"Git Repo {self.alias}: {cmd_name} timed out")
            # Cancel with SIGKILL
            await scmd.cancel(2)

    async def _run_git_cmd(
        self,
        git_args: str,
        timeout: float = 20.,
        attempts: int = 5,
        env: Optional[Dict[str, str]] = None,
        corrupt_hdr: Optional[str] = None,
        log_complete: bool = True
    ) -> str:
        shell_cmd = self.cmd_helper.get_shell_command()
        try:
            return await shell_cmd.exec_cmd(
                f"git -C {self.src_path} {git_args}",
                timeout=timeout,
                attempts=attempts,
                env=env,
                sig_idx=2,
                log_complete=log_complete
            )
        except shell_cmd.error as e:
            stdout = e.stdout.decode().strip()
            stderr = e.stderr.decode().strip()
            msg_lines: List[str] = []
            if stdout:
                msg_lines.extend(stdout.split("\n"))
                self.git_messages.append(stdout)
            if stderr:
                msg_lines.extend(stdout.split("\n"))
                self.git_messages.append(stderr)
            if corrupt_hdr is not None:
                for line in msg_lines:
                    line = line.strip().lower()
                    if line.startswith(corrupt_hdr) and "corrupt" in line:
                        self.repo_corrupt = True
                        break
            raise



# Template Factory helper
#
# Copyright (C) 2021 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.
from __future__ import annotations
import logging
import asyncio
import jinja2
from ..utils import json_wrapper as jsonw
from ..common import RenderableTemplate

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Any,
    Dict
)

if TYPE_CHECKING:
    from ..server import Server
    from ..confighelper import ConfigHelper
    from .secrets import Secrets

class TemplateFactory:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        secrets: Secrets = self.server.load_component(config, 'secrets')
        self.jenv = jinja2.Environment('{%', '%}', '{', '}')
        self.async_env = jinja2.Environment(
            '{%', '%}', '{', '}', enable_async=True
        )
        self.ui_env = jinja2.Environment(enable_async=True)
        self.jenv.add_extension("jinja2.ext.do")
        self.jenv.filters['fromjson'] = jsonw.loads
        self.async_env.add_extension("jinja2.ext.do")
        self.async_env.filters['fromjson'] = jsonw.loads
        self.ui_env.add_extension("jinja2.ext.do")
        self.ui_env.filters['fromjson'] = jsonw.loads
        self.add_environment_global('raise_error', self._raise_error)
        self.add_environment_global('secrets', secrets)

    def add_environment_global(self, name: str, value: Any):
        if name in self.jenv.globals:
            raise self.server.error(
                f"Jinja 2 environment already contains global {name}")
        self.jenv.globals[name] = value
        self.async_env.globals[name] = value

    def _raise_error(self, err_msg: str, err_code: int = 400) -> None:
        raise self.server.error(err_msg, err_code)

    def create_template(self,
                        source: str,
                        is_async: bool = False
                        ) -> JinjaTemplate:
        env = self.async_env if is_async else self.jenv
        try:
            template = env.from_string(source)
        except Exception:
            logging.exception(f"Error creating template from source:\n{source}")
            raise
        return JinjaTemplate(source, self.server, template, is_async)

    def create_ui_template(self, source: str) -> JinjaTemplate:
        try:
            template = self.ui_env.from_string(source)
        except Exception:
            logging.exception(f"Error creating template from source:\n{source}")
            raise
        return JinjaTemplate(source, self.server, template, True)


class JinjaTemplate(RenderableTemplate):
    def __init__(self,
                 source: str,
                 server: Server,
                 template: jinja2.Template,
                 is_async: bool
                 ) -> None:
        self.server = server
        self.orig_source = source
        self.template = template
        self.is_async = is_async

    def __str__(self) -> str:
        return self.orig_source

    def render(self, context: Dict[str, Any] = {}) -> str:
        if self.is_async:
            raise self.server.error(
                "Cannot render async templates with the render() method"
                ", use render_async()")
        try:
            return self.template.render(context).strip()
        except Exception as e:
            msg = "Error rending Jinja2 Template"
            if self.server.is_configured():
                raise self.server.error(msg, 500) from e
            raise self.server.config_error(msg) from e

    async def render_async(self, context: Dict[str, Any] = {}) -> str:
        try:
            ret = await self.template.render_async(context)
        except asyncio.CancelledError:
            raise
        except Exception as e:
            raise self.server.error("Error rending Jinja2 Template", 500) from e
        return ret.strip()

def load_component(config: ConfigHelper) -> TemplateFactory:
    return TemplateFactory(config)



# Common classes used throughout Moonraker
#
# Copyright (C) 2023 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license

from __future__ import annotations
import sys
import logging
import copy
import re
import inspect
import dataclasses
import time
from enum import Enum, Flag, auto
from abc import ABCMeta, abstractmethod
from .utils import ServerError, Sentinel
from .utils import json_wrapper as jsonw

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    Callable,
    Coroutine,
    Type,
    TypeVar,
    Union,
    Dict,
    List,
    Awaitable,
    ClassVar,
    Tuple,
    Generic
)

if TYPE_CHECKING:
    from .server import Server
    from .components.websockets import WebsocketManager
    from .components.authorization import Authorization
    from .components.history import History
    from .components.database import DBProviderWrapper
    from .utils import IPAddress
    from asyncio import Future
    _C = TypeVar("_C", str, bool, float, int)
    _F = TypeVar("_F", bound="ExtendedFlag")
    ConvType = Union[str, bool, float, int]
    ArgVal = Union[None, int, float, bool, str]
    RPCCallback = Callable[..., Coroutine]
    AuthComp = Optional[Authorization]

_T = TypeVar("_T")
ENDPOINT_PREFIXES = ["printer", "server", "machine", "access", "api", "debug"]

class ExtendedFlag(Flag):
    @classmethod
    def from_string(cls: Type[_F], flag_name: str) -> _F:
        str_name = flag_name.upper()
        for name, member in cls.__members__.items():
            if name == str_name:
                return cls(member.value)
        raise ValueError(f"No flag member named {flag_name}")

    @classmethod
    def from_string_list(cls: Type[_F], flag_list: List[str]) -> _F:
        ret = cls(0)
        for flag in flag_list:
            flag = flag.upper()
            ret |= cls.from_string(flag)
        return ret

    @classmethod
    def all(cls: Type[_F]) -> _F:
        return ~cls(0)

    if sys.version_info < (3, 11):
        def __len__(self) -> int:
            return bin(self._value_).count("1")

        def __iter__(self):
            for i in range(self._value_.bit_length()):
                val = 1 << i
                if val & self._value_ == val:
                    yield self.__class__(val)

class RequestType(ExtendedFlag):
    """
    The Request Type is also known as the "Request Method" for
    HTTP/REST APIs.  The use of "Request Method" nomenclature
    is discouraged in Moonraker as it could be confused with
    the JSON-RPC "method" field.
    """
    GET = auto()
    POST = auto()
    DELETE = auto()

class TransportType(ExtendedFlag):
    HTTP = auto()
    WEBSOCKET = auto()
    MQTT = auto()
    INTERNAL = auto()

class ExtendedEnum(Enum):
    @classmethod
    def from_string(cls, enum_name: str):
        str_name = enum_name.upper()
        for name, member in cls.__members__.items():
            if name == str_name:
                return cls(member.value)
        raise ValueError(f"No enum member named {enum_name}")

    def __str__(self) -> str:
        return self._name_.lower()  # type: ignore

class JobEvent(ExtendedEnum):
    STANDBY = 1
    STARTED = 2
    PAUSED = 3
    RESUMED = 4
    COMPLETE = 5
    ERROR = 6
    CANCELLED = 7

    @property
    def finished(self) -> bool:
        return self.value >= 5

    @property
    def aborted(self) -> bool:
        return self.value >= 6

    @property
    def is_printing(self) -> bool:
        return self.value in [2, 4]

class KlippyState(ExtendedEnum):
    DISCONNECTED = 1
    STARTUP = 2
    READY = 3
    ERROR = 4
    SHUTDOWN = 5

    @classmethod
    def from_string(cls, enum_name: str, msg: str = ""):
        str_name = enum_name.upper()
        for name, member in cls.__members__.items():
            if name == str_name:
                instance = cls(member.value)
                if msg:
                    instance.set_message(msg)
                return instance
        raise ValueError(f"No enum member named {enum_name}")


    def set_message(self, msg: str) -> None:
        self._state_message: str = msg

    @property
    def message(self) -> str:
        if hasattr(self, "_state_message"):
            return self._state_message
        return ""

    def startup_complete(self) -> bool:
        return self.value > 2

class RenderableTemplate(metaclass=ABCMeta):
    @abstractmethod
    def __str__(self) -> str:
        ...

    @abstractmethod
    def render(self, context: Dict[str, Any] = {}) -> str:
        ...

    @abstractmethod
    async def render_async(self, context: Dict[str, Any] = {}) -> str:
        ...

@dataclasses.dataclass
class UserInfo:
    username: str
    password: str
    created_on: float = dataclasses.field(default_factory=time.time)
    salt: str = ""
    source: str = "moonraker"
    jwt_secret: Optional[str] = None
    jwk_id: Optional[str] = None
    groups: List[str] = dataclasses.field(default_factory=lambda: ["admin"])

    def as_tuple(self) -> Tuple[Any, ...]:
        return dataclasses.astuple(self)

    def as_dict(self) -> Dict[str, Any]:
        return dataclasses.asdict(self)

@dataclasses.dataclass(frozen=True)
class APIDefinition:
    endpoint: str
    http_path: str
    rpc_methods: List[str]
    request_types: RequestType
    transports: TransportType
    callback: Callable[[WebRequest], Coroutine]
    auth_required: bool
    _cache: ClassVar[Dict[str, APIDefinition]] = {}

    def __str__(self) -> str:
        tprt_str = "|".join([tprt.name for tprt in self.transports if tprt.name])
        val: str = f"(Transports: {tprt_str})"
        if TransportType.HTTP in self.transports:
            req_types = "|".join([rt.name for rt in self.request_types if rt.name])
            val += f" (HTTP Request: {req_types} {self.http_path})"
        if self.rpc_methods:
            methods = " ".join(self.rpc_methods)
            val += f" (RPC Methods: {methods})"
        val += f" (Auth Required: {self.auth_required})"
        return val

    def request(
        self,
        args: Dict[str, Any],
        request_type: RequestType,
        transport: Optional[APITransport] = None,
        ip_addr: Optional[IPAddress] = None,
        user: Optional[UserInfo] = None
    ) -> Coroutine:
        return self.callback(
            WebRequest(self.endpoint, args, request_type, transport, ip_addr, user)
        )

    @property
    def need_object_parser(self) -> bool:
        return self.endpoint.startswith("objects/")

    def rpc_items(self) -> zip[Tuple[RequestType, str]]:
        return zip(self.request_types, self.rpc_methods)

    @classmethod
    def create(
        cls,
        endpoint: str,
        request_types: Union[List[str], RequestType],
        callback: Callable[[WebRequest], Coroutine],
        transports: Union[List[str], TransportType] = TransportType.all(),
        auth_required: bool = True,
        is_remote: bool = False
    ) -> APIDefinition:
        if isinstance(request_types, list):
            request_types = RequestType.from_string_list(request_types)
        if isinstance(transports, list):
            transports = TransportType.from_string_list(transports)
        if endpoint in cls._cache:
            return cls._cache[endpoint]
        http_path = f"/printer/{endpoint.strip('/')}" if is_remote else endpoint
        prf_match = re.match(r"/([^/]+)", http_path)
        if TransportType.HTTP in transports:
            # Validate the first path segment for definitions that support the
            # HTTP transport.  We want to restrict components from registering
            # using unknown paths.
            if prf_match is None or prf_match.group(1) not in ENDPOINT_PREFIXES:
                prefixes = [f"/{prefix} " for prefix in ENDPOINT_PREFIXES]
                raise ServerError(
                    f"Invalid endpoint name '{endpoint}', must start with one of "
                    f"the following: {prefixes}"
                )
        rpc_methods: List[str] = []
        if is_remote:
            # Request Types have no meaning for remote requests.  Therefore
            # both GET and POST http requests are accepted.  JRPC requests do
            # not need an associated RequestType, so the unknown value is used.
            request_types = RequestType.GET | RequestType.POST
            rpc_methods.append(http_path[1:].replace('/', '.'))
        elif transports != TransportType.HTTP:
            name_parts = http_path[1:].split('/')
            if len(request_types) > 1:
                for rtype in request_types:
                    if rtype.name is None:
                        continue
                    func_name = rtype.name.lower() + "_" + name_parts[-1]
                    rpc_methods.append(".".join(name_parts[:-1] + [func_name]))
            else:
                rpc_methods.append(".".join(name_parts))
            if len(request_types) != len(rpc_methods):
                raise ServerError(
                    "Invalid API definition.  Number of websocket methods must "
                    "match the number of request methods"
                )

        api_def = cls(
            endpoint, http_path, rpc_methods, request_types,
            transports, callback, auth_required
        )
        cls._cache[endpoint] = api_def
        return api_def

    @classmethod
    def pop_cached_def(cls, endpoint: str) -> Optional[APIDefinition]:
        return cls._cache.pop(endpoint, None)

    @classmethod
    def get_cache(cls) -> Dict[str, APIDefinition]:
        return cls._cache

    @classmethod
    def reset_cache(cls) -> None:
        cls._cache.clear()

class APITransport:
    @property
    def transport_type(self) -> TransportType:
        return TransportType.INTERNAL

    @property
    def user_info(self) -> Optional[UserInfo]:
        return None

    @property
    def ip_addr(self) -> Optional[IPAddress]:
        return None

    def screen_rpc_request(
        self, api_def: APIDefinition, req_type: RequestType, args: Dict[str, Any]
    ) -> None:
        return None

    def send_status(
        self, status: Dict[str, Any], eventtime: float
    ) -> None:
        raise NotImplementedError

class BaseRemoteConnection(APITransport):
    def on_create(self, server: Server) -> None:
        self.server = server
        self.eventloop = server.get_event_loop()
        self.wsm: WebsocketManager = self.server.lookup_component("websockets")
        self.rpc: JsonRPC = self.server.lookup_component("jsonrpc")
        self._uid = id(self)
        self.is_closed: bool = False
        self.queue_busy: bool = False
        self.pending_responses: Dict[int, Future] = {}
        self.message_buf: List[Union[bytes, str]] = []
        self._connected_time: float = 0.
        self._identified: bool = False
        self._client_data: Dict[str, str] = {
            "name": "unknown",
            "version": "",
            "type": "",
            "url": ""
        }
        self._need_auth: bool = False
        self._user_info: Optional[UserInfo] = None

    @property
    def user_info(self) -> Optional[UserInfo]:
        return self._user_info

    @user_info.setter
    def user_info(self, uinfo: UserInfo) -> None:
        self._user_info = uinfo
        self._need_auth = False

    @property
    def need_auth(self) -> bool:
        return self._need_auth

    @property
    def uid(self) -> int:
        return self._uid

    @property
    def hostname(self) -> str:
        return ""

    @property
    def start_time(self) -> float:
        return self._connected_time

    @property
    def identified(self) -> bool:
        return self._identified

    @property
    def client_data(self) -> Dict[str, str]:
        return self._client_data

    @client_data.setter
    def client_data(self, data: Dict[str, str]) -> None:
        self._client_data = data
        self._identified = True

    @property
    def transport_type(self) -> TransportType:
        return TransportType.WEBSOCKET

    def screen_rpc_request(
        self, api_def: APIDefinition, req_type: RequestType, args: Dict[str, Any]
    ) -> None:
        self.check_authenticated(api_def)

    async def _process_message(self, message: str) -> None:
        try:
            response = await self.rpc.dispatch(message, self)
            if response is not None:
                self.queue_message(response)
        except Exception:
            logging.exception("Websocket Command Error")

    def queue_message(self, message: Union[bytes, str, Dict[str, Any]]):
        self.message_buf.append(
            jsonw.dumps(message) if isinstance(message, dict) else message
        )
        if self.queue_busy:
            return
        self.queue_busy = True
        self.eventloop.register_callback(self._write_messages)

    def authenticate(
        self,
        token: Optional[str] = None,
        api_key: Optional[str] = None
    ) -> None:
        auth: AuthComp = self.server.lookup_component("authorization", None)
        if auth is None:
            return
        if token is not None:
            self.user_info = auth.validate_jwt(token)
        elif api_key is not None and self.user_info is None:
            self.user_info = auth.validate_api_key(api_key)
        elif self._need_auth:
            raise self.server.error("Unauthorized", 401)

    def check_authenticated(self, api_def: APIDefinition) -> None:
        if not self._need_auth:
            return
        auth: AuthComp = self.server.lookup_component("authorization", None)
        if auth is None:
            return
        if api_def.auth_required:
            raise self.server.error("Unauthorized", 401)

    def on_user_logout(self, user: str) -> bool:
        if self._user_info is None:
            return False
        if user == self._user_info.username:
            self._user_info = None
            return True
        return False

    async def _write_messages(self):
        if self.is_closed:
            self.message_buf = []
            self.queue_busy = False
            return
        while self.message_buf:
            msg = self.message_buf.pop(0)
            await self.write_to_socket(msg)
        self.queue_busy = False

    async def write_to_socket(self, message: Union[bytes, str]) -> None:
        raise NotImplementedError("Children must implement write_to_socket")

    def send_status(self,
                    status: Dict[str, Any],
                    eventtime: float
                    ) -> None:
        if not status:
            return
        self.queue_message({
            'jsonrpc': "2.0",
            'method': "notify_status_update",
            'params': [status, eventtime]})

    def call_method_with_response(
        self,
        method: str,
        params: Optional[Union[List, Dict[str, Any]]] = None,
    ) -> Awaitable:
        fut = self.eventloop.create_future()
        msg: Dict[str, Any] = {
            'jsonrpc': "2.0",
            'method': method,
            'id': id(fut)
        }
        if params:
            msg["params"] = params
        self.pending_responses[id(fut)] = fut
        self.queue_message(msg)
        return fut

    def call_method(
        self,
        method: str,
        params: Optional[Union[List, Dict[str, Any]]] = None
    ) -> None:
        msg: Dict[str, Any] = {
            "jsonrpc": "2.0",
            "method": method
        }
        if params:
            msg["params"] = params
        self.queue_message(msg)

    def send_notification(self, name: str, data: List) -> None:
        self.wsm.notify_clients(name, data, [self._uid])

    def resolve_pending_response(
        self, response_id: int, result: Any
    ) -> bool:
        fut = self.pending_responses.pop(response_id, None)
        if fut is None:
            return False
        if isinstance(result, ServerError):
            fut.set_exception(result)
        else:
            fut.set_result(result)
        return True

    def close_socket(self, code: int, reason: str) -> None:
        raise NotImplementedError("Children must implement close_socket()")


class WebRequest:
    def __init__(
        self,
        endpoint: str,
        args: Dict[str, Any],
        request_type: RequestType = RequestType(0),
        transport: Optional[APITransport] = None,
        ip_addr: Optional[IPAddress] = None,
        user: Optional[UserInfo] = None
    ) -> None:
        self.endpoint = endpoint
        self.args = args
        self.transport = transport
        self.request_type = request_type
        self.ip_addr: Optional[IPAddress] = ip_addr
        self.current_user = user

    def get_endpoint(self) -> str:
        return self.endpoint

    def get_request_type(self) -> RequestType:
        return self.request_type

    def get_action(self) -> str:
        return self.request_type.name or ""

    def get_args(self) -> Dict[str, Any]:
        return self.args

    def get_subscribable(self) -> Optional[APITransport]:
        return self.transport

    def get_client_connection(self) -> Optional[BaseRemoteConnection]:
        if isinstance(self.transport, BaseRemoteConnection):
            return self.transport
        return None

    def get_ip_address(self) -> Optional[IPAddress]:
        return self.ip_addr

    def get_current_user(self) -> Optional[UserInfo]:
        return self.current_user

    def _get_converted_arg(self,
                           key: str,
                           default: Union[Sentinel, _T],
                           dtype: Type[_C]
                           ) -> Union[_C, _T]:
        if key not in self.args:
            if default is Sentinel.MISSING:
                raise ServerError(f"No data for argument: {key}")
            return default
        val = self.args[key]
        try:
            if dtype is not bool:
                return dtype(val)
            else:
                if isinstance(val, str):
                    val = val.lower()
                    if val in ["true", "false"]:
                        return True if val == "true" else False  # type: ignore
                elif isinstance(val, bool):
                    return val  # type: ignore
                raise TypeError
        except Exception:
            raise ServerError(
                f"Unable to convert argument [{key}] to {dtype}: "
                f"value recieved: {val}")

    def get(self,
            key: str,
            default: Union[Sentinel, _T] = Sentinel.MISSING
            ) -> Union[_T, Any]:
        val = self.args.get(key, default)
        if val is Sentinel.MISSING:
            raise ServerError(f"No data for argument: {key}")
        return val

    def get_str(self,
                key: str,
                default: Union[Sentinel, _T] = Sentinel.MISSING
                ) -> Union[str, _T]:
        return self._get_converted_arg(key, default, str)

    def get_int(self,
                key: str,
                default: Union[Sentinel, _T] = Sentinel.MISSING
                ) -> Union[int, _T]:
        return self._get_converted_arg(key, default, int)

    def get_float(self,
                  key: str,
                  default: Union[Sentinel, _T] = Sentinel.MISSING
                  ) -> Union[float, _T]:
        return self._get_converted_arg(key, default, float)

    def get_boolean(self,
                    key: str,
                    default: Union[Sentinel, _T] = Sentinel.MISSING
                    ) -> Union[bool, _T]:
        return self._get_converted_arg(key, default, bool)

    def _parse_list(
        self,
        key: str,
        sep: str,
        ltype: Type[_C],
        count: Optional[int],
        default: Union[Sentinel, _T]
    ) -> Union[List[_C], _T]:
        if key not in self.args:
            if default is Sentinel.MISSING:
                raise ServerError(f"No data for argument: {key}")
            return default
        value = self.args[key]
        if isinstance(value, str):
            try:
                ret = [ltype(val.strip()) for val in value.split(sep) if val.strip()]
            except Exception as e:
                raise ServerError(
                    f"Invalid list format received for argument '{key}', "
                    "parsing failed."
                ) from e
        elif isinstance(value, list):
            for val in value:
                if not isinstance(val, ltype):
                    raise ServerError(
                        f"Invalid list format for argument '{key}', expected all "
                        f"values to be of type {ltype.__name__}."
                    )
            # List already parsed
            ret = value
        else:
            raise ServerError(
                f"Invalid value received for argument '{key}'.  Expected List type, "
                f"received {type(value).__name__}"
            )
        if count is not None and len(ret) != count:
            raise ServerError(
                f"Invalid list received for argument '{key}', count mismatch. "
                f"Expected {count} items, got {len(ret)}."
            )
        return ret

    def get_list(
        self,
        key: str,
        default: Union[Sentinel, _T] = Sentinel.MISSING,
        sep: str = ",",
        count: Optional[int] = None
    ) -> Union[_T, List[str]]:
        return self._parse_list(key, sep, str, count, default)


class JsonRPC:
    def __init__(self, server: Server) -> None:
        self.methods: Dict[str, Tuple[RequestType, APIDefinition]] = {}
        self.sanitize_response = False
        self.verbose = server.is_verbose_enabled()

    def _log_request(self, rpc_obj: Dict[str, Any], trtype: TransportType) -> None:
        if not self.verbose:
            return
        self.sanitize_response = False
        output = rpc_obj
        method: Optional[str] = rpc_obj.get("method")
        params: Dict[str, Any] = rpc_obj.get("params", {})
        if isinstance(method, str):
            if (
                method.startswith("access.") or
                method == "machine.sudo.password"
            ):
                self.sanitize_response = True
                if params and isinstance(params, dict):
                    output = copy.deepcopy(rpc_obj)
                    output["params"] = {key: "<sanitized>" for key in params}
            elif method == "server.connection.identify":
                output = copy.deepcopy(rpc_obj)
                for field in ["access_token", "api_key"]:
                    if field in params:
                        output["params"][field] = "<sanitized>"
        logging.debug(f"{trtype} Received::{jsonw.dumps(output).decode()}")

    def _log_response(
        self, resp_obj: Optional[Dict[str, Any]], trtype: TransportType
    ) -> None:
        if not self.verbose:
            return
        if resp_obj is None:
            return
        output = resp_obj
        if self.sanitize_response and "result" in resp_obj:
            output = copy.deepcopy(resp_obj)
            output["result"] = "<sanitized>"
        self.sanitize_response = False
        logging.debug(f"{trtype} Response::{jsonw.dumps(output).decode()}")

    def register_method(
        self,
        name: str,
        request_type: RequestType,
        api_definition: APIDefinition
    ) -> None:
        self.methods[name] = (request_type, api_definition)

    def get_method(self, name: str) -> Optional[Tuple[RequestType, APIDefinition]]:
        return self.methods.get(name, None)

    def remove_method(self, name: str) -> None:
        self.methods.pop(name, None)

    async def dispatch(
        self,
        data: Union[str, bytes],
        transport: APITransport
    ) -> Optional[bytes]:
        transport_type = transport.transport_type
        try:
            obj: Union[Dict[str, Any], List[dict]] = jsonw.loads(data)
        except Exception:
            if isinstance(data, bytes):
                data = data.decode()
            msg = f"{transport_type} data not valid json: {data}"
            logging.exception(msg)
            err = self.build_error(-32700, "Parse error")
            return jsonw.dumps(err)
        if isinstance(obj, list):
            responses: List[Dict[str, Any]] = []
            for item in obj:
                self._log_request(item, transport_type)
                resp = await self.process_object(item, transport)
                if resp is not None:
                    self._log_response(resp, transport_type)
                    responses.append(resp)
            if responses:
                return jsonw.dumps(responses)
        else:
            self._log_request(obj, transport_type)
            response = await self.process_object(obj, transport)
            if response is not None:
                self._log_response(response, transport_type)
                return jsonw.dumps(response)
        return None

    async def process_object(
        self,
        obj: Dict[str, Any],
        transport: APITransport
    ) -> Optional[Dict[str, Any]]:
        req_id: Optional[int] = obj.get('id', None)
        rpc_version: str = obj.get('jsonrpc', "")
        if rpc_version != "2.0":
            return self.build_error(-32600, "Invalid Request", req_id)
        method_name = obj.get('method', Sentinel.MISSING)
        if method_name is Sentinel.MISSING:
            self.process_response(obj, transport)
            return None
        if not isinstance(method_name, str):
            return self.build_error(
                -32600, "Invalid Request", req_id, method_name=str(method_name)
            )
        method_info = self.methods.get(method_name, None)
        if method_info is None:
            return self.build_error(
                -32601, "Method not found", req_id, method_name=method_name
            )
        request_type, api_definition = method_info
        transport_type = transport.transport_type
        if transport_type not in api_definition.transports:
            return self.build_error(
                -32601, f"Method not found for transport {transport_type.name}",
                req_id, method_name=method_name
            )
        params: Dict[str, Any] = {}
        if 'params' in obj:
            params = obj['params']
            if not isinstance(params, dict):
                return self.build_error(
                    -32602, "Invalid params:", req_id, method_name=method_name
                )
        return await self.execute_method(
            method_name, request_type, api_definition, req_id, transport, params
        )

    def process_response(
        self, obj: Dict[str, Any], conn: APITransport
    ) -> None:
        if not isinstance(conn, BaseRemoteConnection):
            logging.debug(f"RPC Response to non-socket request: {obj}")
            return
        response_id = obj.get("id")
        if response_id is None:
            logging.debug(f"RPC Response with null ID: {obj}")
            return
        result = obj.get("result")
        if result is None:
            name = conn.client_data["name"]
            error = obj.get("error")
            msg = f"Invalid Response: {obj}"
            code = -32600
            if isinstance(error, dict):
                msg = error.get("message", msg)
                code = error.get("code", code)
            msg = f"{name} rpc error: {code} {msg}"
            ret = ServerError(msg, 418)
        else:
            ret = result
        conn.resolve_pending_response(response_id, ret)

    async def execute_method(
        self,
        method_name: str,
        request_type: RequestType,
        api_definition: APIDefinition,
        req_id: Optional[int],
        transport: APITransport,
        params: Dict[str, Any]
    ) -> Optional[Dict[str, Any]]:
        try:
            transport.screen_rpc_request(api_definition, request_type, params)
            result = await api_definition.request(
                params, request_type, transport, transport.ip_addr, transport.user_info
            )
        except TypeError as e:
            return self.build_error(
                -32602, f"Invalid params:\n{e}", req_id, True, method_name
            )
        except ServerError as e:
            code = e.status_code
            if code == 404:
                code = -32601
            elif code == 401:
                code = -32602
            return self.build_error(code, str(e), req_id, True, method_name)
        except Exception as e:
            return self.build_error(-31000, str(e), req_id, True, method_name)

        if req_id is None:
            return None
        else:
            return self.build_result(result, req_id)

    def build_result(self, result: Any, req_id: int) -> Dict[str, Any]:
        return {
            'jsonrpc': "2.0",
            'result': result,
            'id': req_id
        }

    def build_error(
        self,
        code: int,
        msg: str,
        req_id: Optional[int] = None,
        is_exc: bool = False,
        method_name: str = ""
    ) -> Dict[str, Any]:
        if method_name:
            method_name = f"Requested Method: {method_name}, "
        log_msg = f"JSON-RPC Request Error - {method_name}Code: {code}, Message: {msg}"
        if is_exc and self.verbose:
            logging.exception(log_msg)
        else:
            logging.info(log_msg)
        return {
            'jsonrpc': "2.0",
            'error': {'code': code, 'message': msg},
            'id': req_id
        }


# *** Job History Common Clases ***

class FieldTracker(Generic[_T]):
    history: History = None  # type: ignore
    def __init__(
        self,
        value: _T = None,  # type: ignore
        reset_callback: Optional[Callable[[], _T]] = None,
        exclude_paused: bool = False,
    ) -> None:
        self.tracked_value = value
        self.exclude_paused = exclude_paused
        self.reset_callback: Optional[Callable[[], _T]] = reset_callback

    def set_reset_callback(self, cb: Optional[Callable[[], _T]]) -> None:
        self.reset_callback = cb

    def set_exclude_paused(self, exclude: bool) -> None:
        self.exclude_paused = exclude

    def reset(self) -> None:
        raise NotImplementedError()

    def update(self, value: _T) -> None:
        raise NotImplementedError()

    def get_tracked_value(self) -> _T:
        return self.tracked_value

    def has_totals(self) -> bool:
        return False

    @classmethod
    def class_init(cls, history: History) -> None:
        cls.history = history


class BasicTracker(FieldTracker[Any]):
    def __init__(
        self,
        value: Any = None,
        reset_callback: Optional[Callable[[], Any]] = None,
        exclude_paused: bool = False
    ) -> None:
        super().__init__(value, reset_callback, exclude_paused)

    def reset(self) -> None:
        if self.reset_callback is not None:
            self.tracked_value = self.reset_callback()

    def update(self, value: Any) -> None:
        if self.history.tracking_enabled(self.exclude_paused):
            self.tracked_value = value

    def has_totals(self) -> bool:
        return isinstance(self.tracked_value, (int, float))


class DeltaTracker(FieldTracker[Union[int, float]]):
    def __init__(
        self,
        value: Union[int, float] = 0,
        reset_callback: Optional[Callable[[], Union[float, int]]] = None,
        exclude_paused: bool = False
    ) -> None:
        super().__init__(value, reset_callback, exclude_paused)
        self.last_value: Union[float, int, None] = None

    def reset(self) -> None:
        self.tracked_value = 0
        self.last_value = None
        if self.reset_callback is not None:
            self.last_value = self.reset_callback()
            if not isinstance(self.last_value, (float, int)):
                logging.info("DeltaTracker reset to invalid type")
                self.last_value = None

    def update(self, value: Union[int, float]) -> None:
        if not isinstance(value, (int, float)):
            return
        if self.history.tracking_enabled(self.exclude_paused):
            if self.last_value is not None:
                self.tracked_value += value - self.last_value
        self.last_value = value

    def has_totals(self) -> bool:
        return True


class CumulativeTracker(FieldTracker[Union[int, float]]):
    def __init__(
        self,
        value: Union[int, float] = 0,
        reset_callback: Optional[Callable[[], Union[float, int]]] = None,
        exclude_paused: bool = False
    ) -> None:
        super().__init__(value, reset_callback, exclude_paused)

    def reset(self) -> None:
        if self.reset_callback is not None:
            self.tracked_value = self.reset_callback()
            if not isinstance(self.tracked_value, (float, int)):
                logging.info(f"{self.__class__.__name__} reset to invalid type")
                self.tracked_value = 0
        else:
            self.tracked_value = 0

    def update(self, value: Union[int, float]) -> None:
        if not isinstance(value, (int, float)):
            return
        if self.history.tracking_enabled(self.exclude_paused):
            self.tracked_value += value

    def has_totals(self) -> bool:
        return True

class AveragingTracker(CumulativeTracker):
    def __init__(
        self,
        value: Union[int, float] = 0,
        reset_callback: Optional[Callable[[], Union[float, int]]] = None,
        exclude_paused: bool = False
    ) -> None:
        super().__init__(value, reset_callback, exclude_paused)
        self.count = 0

    def reset(self) -> None:
        super().reset()
        self.count = 0

    def update(self, value: Union[int, float]) -> None:
        if not isinstance(value, (int, float)):
            return
        if self.history.tracking_enabled(self.exclude_paused):
            lv = self.tracked_value
            self.count += 1
            self.tracked_value = (lv * (self.count - 1) + value) / self.count


class MaximumTracker(CumulativeTracker):
    def __init__(
        self,
        value: Union[int, float] = 0,
        reset_callback: Optional[Callable[[], Union[float, int]]] = None,
        exclude_paused: bool = False
    ) -> None:
        super().__init__(value, reset_callback, exclude_paused)
        self.initialized = False

    def reset(self) -> None:
        self.initialized = False
        if self.reset_callback is not None:
            self.tracked_value = self.reset_callback()
            if not isinstance(self.tracked_value, (int, float)):
                self.tracked_value = 0
                logging.info("MaximumTracker reset to invalid type")
            else:
                self.initialized = True
        else:
            self.tracked_value = 0

    def update(self, value: Union[float, int]) -> None:
        if not isinstance(value, (int, float)):
            return
        if self.history.tracking_enabled(self.exclude_paused):
            if not self.initialized:
                self.tracked_value = value
                self.initialized = True
            else:
                self.tracked_value = max(self.tracked_value, value)

class MinimumTracker(CumulativeTracker):
    def __init__(
        self,
        value: Union[int, float] = 0,
        reset_callback: Optional[Callable[[], Union[float, int]]] = None,
        exclude_paused: bool = False
    ) -> None:
        super().__init__(value, reset_callback, exclude_paused)
        self.initialized = False

    def reset(self) -> None:
        self.initialized = False
        if self.reset_callback is not None:
            self.tracked_value = self.reset_callback()
            if not isinstance(self.tracked_value, (int, float)):
                self.tracked_value = 0
                logging.info("MinimumTracker reset to invalid type")
            else:
                self.initialized = True
        else:
            self.tracked_value = 0

    def update(self, value: Union[float, int]) -> None:
        if not isinstance(value, (int, float)):
            return
        if self.history.tracking_enabled(self.exclude_paused):
            if not self.initialized:
                self.tracked_value = value
                self.initialized = True
            else:
                self.tracked_value = min(self.tracked_value, value)

class CollectionTracker(FieldTracker[List[Any]]):
    MAX_SIZE = 100
    def __init__(
        self,
        value: List[Any] = [],
        reset_callback: Optional[Callable[[], List[Any]]] = None,
        exclude_paused: bool = False
    ) -> None:
        super().__init__(list(value), reset_callback, exclude_paused)

    def reset(self) -> None:
        if self.reset_callback is not None:
            self.tracked_value = self.reset_callback()
            if not isinstance(self.tracked_value, list):
                logging.info("CollectionTracker reset to invalid type")
                self.tracked_value = []
        else:
            self.tracked_value.clear()

    def update(self, value: Any) -> None:
        if value in self.tracked_value:
            return
        if self.history.tracking_enabled(self.exclude_paused):
            self.tracked_value.append(value)
            if len(self.tracked_value) > self.MAX_SIZE:
                self.tracked_value.pop(0)

    def has_totals(self) -> bool:
        return False


class TrackingStrategy(ExtendedEnum):
    BASIC = 1
    DELTA = 2
    ACCUMULATE = 3
    AVERAGE = 4
    MAXIMUM = 5
    MINIMUM = 6
    COLLECT = 7

    def get_tracker(self, **kwargs) -> FieldTracker:
        trackers: Dict[TrackingStrategy, Type[FieldTracker]] = {
            TrackingStrategy.BASIC: BasicTracker,
            TrackingStrategy.DELTA: DeltaTracker,
            TrackingStrategy.ACCUMULATE: CumulativeTracker,
            TrackingStrategy.AVERAGE: AveragingTracker,
            TrackingStrategy.MAXIMUM: MaximumTracker,
            TrackingStrategy.MINIMUM: MinimumTracker,
            TrackingStrategy.COLLECT: CollectionTracker
        }
        return trackers[self](**kwargs)


class HistoryFieldData:
    def __init__(
        self,
        field_name: str,
        provider: str,
        desc: str,
        strategy: str,
        units: Optional[str] = None,
        reset_callback: Optional[Callable[[], _T]] = None,
        exclude_paused: bool = False,
        report_total: bool = False,
        report_maximum: bool = False,
        precision: Optional[int] = None
    ) -> None:
        self._name = field_name
        self._provider = provider
        self._desc = desc
        self._strategy = TrackingStrategy.from_string(strategy)
        self._units = units
        self._tracker = self._strategy.get_tracker(
            reset_callback=reset_callback,
            exclude_paused=exclude_paused
        )
        self._report_total = report_total
        self._report_maximum = report_maximum
        self._precision = precision

    @property
    def name(self) -> str:
        return self._name

    @property
    def provider(self) -> str:
        return self._provider

    @property
    def tracker(self) -> FieldTracker:
        return self._tracker

    def __eq__(self, value: object) -> bool:
        if isinstance(value, HistoryFieldData):
            return value._provider == self._provider and value._name == self._name
        raise ValueError("Invalid type for comparison")

    def get_configuration(self) -> Dict[str, Any]:
        return {
            "field": self._name,
            "provider": self._provider,
            "description": self._desc,
            "strategy": self._strategy.name.lower(),
            "units": self._units,
            "init_tracker": self._tracker.reset_callback is not None,
            "exclude_paused": self._tracker.exclude_paused,
            "report_total": self._report_total,
            "report_maximum": self._report_maximum,
            "precision": self._precision
        }

    def as_dict(self) -> Dict[str, Any]:
        val = self._tracker.get_tracked_value()
        if self._precision is not None and isinstance(val, float):
            val = round(val, self._precision)
        return {
            "provider": self._provider,
            "name": self.name,
            "value": val,
            "description": self._desc,
            "units": self._units
        }

    def has_totals(self) -> bool:
        return (
            self._tracker.has_totals() and
            (self._report_total or self._report_maximum)
        )

    def get_totals(
        self, last_totals: List[Dict[str, Any]], reset: bool = False
    ) -> Dict[str, Any]:
        if not self.has_totals():
            return {}
        if reset:
            maximum: Optional[float] = 0 if self._report_maximum else None
            total: Optional[float] = 0 if self._report_total else None
        else:
            cur_val: Union[float, int] = self._tracker.get_tracked_value()
            maximum = cur_val if self._report_maximum else None
            total = cur_val if self._report_total else None
            for obj in last_totals:
                if obj["provider"] == self._provider and obj["field"] == self._name:
                    if maximum is not None:
                        maximum = max(cur_val, obj["maximum"] or 0)
                    if total is not None:
                        total = cur_val + (obj["total"] or 0)
                    break
            if self._precision is not None:
                if maximum is not None:
                    maximum = round(maximum, self._precision)
                if total is not None:
                    total = round(total, self._precision)
        return {
            "provider": self._provider,
            "field": self._name,
            "maximum": maximum,
            "total": total
        }

class SqlTableDefType(type):
    def __new__(
        metacls,
        clsname: str,
        bases: Tuple[type, ...],
        cls_attrs: Dict[str, Any]
    ):
        if clsname != "SqlTableDefinition":
            for item in ("name", "prototype"):
                if not cls_attrs[item]:
                    raise ValueError(
                        f"Class attribute `{item}` must be set for class {clsname}"
                    )
            if cls_attrs["version"] < 1:
                raise ValueError(
                    f"The 'version' attribute of {clsname} must be greater than 0"
                )
            cls_attrs["prototype"] = inspect.cleandoc(cls_attrs["prototype"].strip())
            prototype = cls_attrs["prototype"]
            proto_match = re.match(
                r"([a-zA-Z][0-9a-zA-Z_-]+)\s*\((.+)\)\s*;?$", prototype, re.DOTALL
            )
            if proto_match is None:
                raise ValueError(f"Invalid SQL Table prototype:\n{prototype}")
            table_name = cls_attrs["name"]
            parsed_name = proto_match.group(1)
            if table_name != parsed_name:
                raise ValueError(
                    f"Table name '{table_name}' does not match parsed name from "
                    f"table prototype '{parsed_name}'"
                )
        return super().__new__(metacls, clsname, bases, cls_attrs)

class SqlTableDefinition(metaclass=SqlTableDefType):
    name: str = ""
    version: int = 0
    prototype: str = ""
    def __init__(self) -> None:
        if self.__class__ == SqlTableDefinition:
            raise ServerError("Cannot directly instantiate SqlTableDefinition")

    def migrate(
        self, last_version: int, db_provider: DBProviderWrapper
    ) -> None:
        raise NotImplementedError("Children must implement migrate")



# WLED neopixel support
#
# Copyright (C) 2021-2022 Richard Mitchell <richardjm+moonraker@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.

# Component to control the wled neopixel home system from AirCookie
# Github at https://github.com/Aircoookie/WLED
# Wiki at https://kno.wled.ge/

from __future__ import annotations
from enum import Enum
import logging
import asyncio
import serial_asyncio
from tornado.httpclient import AsyncHTTPClient
from tornado.httpclient import HTTPRequest
from ..utils import json_wrapper as jsonw
from ..common import RequestType

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Type,
    List,
    Any,
    Optional,
    Dict,
)

if TYPE_CHECKING:
    from ..confighelper import ConfigHelper
    from ..common import WebRequest

class OnOff(str, Enum):
    on: str = "on"
    off: str = "off"

class Strip():
    _COLORSIZE: int = 4

    def __init__(self: Strip,
                 name: str,
                 cfg: ConfigHelper):
        self.server = cfg.get_server()
        self.request_mutex = asyncio.Lock()

        self.name = name

        self.initial_preset: int = cfg.getint("initial_preset", -1)
        self.initial_red: float = cfg.getfloat("initial_red", 0.5)
        self.initial_green: float = cfg.getfloat("initial_green", 0.5)
        self.initial_blue: float = cfg.getfloat("initial_blue", 0.5)
        self.initial_white: float = cfg.getfloat("initial_white", 0.5)
        self.chain_count: int = cfg.getint("chain_count", 1)

        # Supports rgbw always
        self._chain_data = bytearray(
            self.chain_count * self._COLORSIZE)

        self.onoff = OnOff.off
        self.preset = self.initial_preset

    def get_strip_info(self: Strip) -> Dict[str, Any]:
        return {
            "strip": self.name,
            "status": self.onoff.value,
            "chain_count": self.chain_count,
            "preset": self.preset,
            "brightness": self.brightness,
            "intensity": self.intensity,
            "speed": self.speed,
            "error": self.error_state
        }

    async def initialize(self: Strip) -> None:
        self.send_full_chain_data = True
        self.onoff = OnOff.on
        self.preset = self.initial_preset
        self.brightness = 255
        self.intensity = -1
        self.speed = -1
        if self.initial_preset >= 0:
            self._update_color_data(self.initial_red,
                                    self.initial_green,
                                    self.initial_blue,
                                    self.initial_white,
                                    None)
            await self.wled_on(self.initial_preset)
        else:
            await self.set_wled(self.initial_red,
                                self.initial_green,
                                self.initial_blue,
                                self.initial_white,
                                None,
                                True)

    def _update_color_data(self: Strip,
                           red: float, green: float, blue: float, white: float,
                           index: Optional[int]) -> None:
        red = int(red * 255. + .5)
        blue = int(blue * 255. + .5)
        green = int(green * 255. + .5)
        white = int(white * 255. + .5)
        led_data = [red, green, blue, white]

        if index is None:
            self._chain_data[:] = led_data * self.chain_count
        else:
            elem_size = len(led_data)
            self._chain_data[(index-1)*elem_size:index*elem_size] = led_data

    async def send_wled_command_impl(self: Strip,
                                     state: Dict[str, Any]) -> None:
        pass

    def close(self: Strip):
        pass

    async def _send_wled_command(self: Strip,
                                 state: Dict[str, Any]) -> None:
        try:
            await self.send_wled_command_impl(state)

            self.error_state = None
        except Exception as e:
            msg = f"WLED: Error {e}"
            self.error_state = msg
            logging.exception(msg)
            raise self.server.error(msg)

    async def wled_on(self: Strip, preset: int) -> None:
        self.onoff = OnOff.on
        logging.debug(f"WLED: {self.name} on PRESET={preset}")
        if preset < 0:
            # WLED_ON STRIP=strip (no args) - reset to default
            await self.initialize()
        else:
            self.send_full_chain_data = True
            self.preset = preset
            # Without reading the data back from wled we don't know the values
            self.brightness = -1
            self.intensity = -1
            self.speed = -1
            await self._send_wled_command({"on": True, "ps": preset})

    async def wled_off(self: Strip) -> None:
        logging.debug(f"WLED: {self.name} off")
        self.onoff = OnOff.off
        # Without this calling SET_WLED for a single pixel after WLED_OFF
        # would send just that pixel
        self.send_full_chain_data = True
        await self._send_wled_command({"on": False})

    async def wled_control(self: Strip, brightness: int, intensity: int,
                           speed: int) -> None:
        logging.debug(
            f"WLED: {self.name} control {self.onoff} BRIGHTNESS={brightness} "
            f"INTENSITY={intensity} SPEED={speed} CURRENTPRESET={self.preset}")

        if self.onoff == OnOff.off:
            logging.info("wled control only permitted when strip is on")
            return

        # Even if a preset is not activated sending seg {} information will
        # turn it back on
        control: Dict[str, Any]
        if self.preset != -1:
            control = {"tt": 0, "seg": {}}
        else:
            control = {"tt": 0}

        shouldSend: bool = False
        # Using 0 is not recommended in wled docs
        if brightness > 0:
            if brightness > 255:
                logging.info("BRIGHTNESS should be between 1 and 255")
            else:
                shouldSend = True
                self.brightness = brightness
                control["bri"] = self.brightness
                # Brightness in seg {} - only if a preset is on
                if self.preset != -1:
                    control["seg"]["bri"] = self.brightness

        # Intensity - only if a preset is on
        if intensity > -1 and self.preset != -1:
            if intensity > 255:
                logging.info("INTENSITY should be between 0 and 255")
            else:
                shouldSend = True
                self.intensity = intensity
                control["seg"]["ix"] = self.intensity

        # Speed - only if a preset is on
        if speed > -1 and self.preset != -1:
            if speed > 255:
                logging.info("SPEED should be between 0 and 255")
            else:
                shouldSend = True
                self.speed = speed
                control["seg"]["sx"] = self.speed

        # Control brightness, intensity, and speed for segment
        # This will allow full control for effects such as "Percent"
        if shouldSend:
            await self._send_wled_command(control)

    def _wled_pixel(self: Strip, index: int) -> List[int]:
        led_color_data: List[int] = []
        for p in self._chain_data[(index-1)*self._COLORSIZE:
                                  (index)*self._COLORSIZE]:
            led_color_data.append(p)
        return led_color_data

    async def set_wled(self: Strip,
                       red: float, green: float, blue: float, white: float,
                       index: Optional[int], transmit: bool) -> None:
        logging.debug(
            f"WLED: {self.name} R={red} G={green} B={blue} W={white} "
            f"INDEX={index} TRANSMIT={transmit}")
        self._update_color_data(red, green, blue, white, index)
        if transmit:
            # Clear preset (issues with sending seg{} will revert to preset)
            self.preset = -1

            # If we are coming from a preset without a wled_control
            # we don't know a brightness, this will also ensure
            # behaviour is consistent prior to introduction of wled_control
            if self.brightness == -1:
                self.brightness = 255

            # Base command for setting an led (for all active segments)
            # See https://kno.wled.ge/interfaces/json-api/
            state: Dict[str, Any] = {"on": True,
                                     "tt": 0,
                                     "bri": self.brightness,
                                     "seg": {"bri": self.brightness, "i": []}}
            if index is None:
                # All pixels same color only send range command of first color
                self.send_full_chain_data = False
                state["seg"]["i"] = [0, self.chain_count, self._wled_pixel(1)]
            elif self.send_full_chain_data:
                # Send a full set of color data (e.g. previous preset)
                self.send_full_chain_data = False
                cdata = []
                for i in range(self.chain_count):
                    cdata.append(self._wled_pixel(i+1))
                state["seg"]["i"] = cdata
            else:
                # Only one pixel has changed since last full data sent
                # so send just that one
                state["seg"]["i"] = [index-1, self._wled_pixel(index)]

            # Send wled control command
            await self._send_wled_command(state)

            if self.onoff == OnOff.off:
                # Without a repeated call individual led control doesn't
                # turn the led strip back on or doesn't set brightness
                # correctly from off
                # Confirmed as a bug:
                # https://discord.com/channels/473448917040758787/757254961640898622/934135556370202645
                self.onoff = OnOff.on
                await self._send_wled_command(state)
        else:
            # If not transmitting this time easiest just to send all data when
            # next transmitting
            self.send_full_chain_data = True

class StripHttp(Strip):
    def __init__(self: StripHttp,
                 name: str,
                 cfg: ConfigHelper):
        super().__init__(name, cfg)

        # Read the uri information
        addr: str = cfg.get("address")
        port: int = cfg.getint("port", 80)
        protocol: str = cfg.get("protocol", "http")
        self.url = f"{protocol}://{addr}:{port}/json"

        self.timeout: float = cfg.getfloat("timeout", 2.)
        self.client = AsyncHTTPClient()

    async def send_wled_command_impl(self: StripHttp,
                                     state: Dict[str, Any],
                                     retries: int = 3
                                     ) -> None:
        async with self.request_mutex:
            logging.debug(f"WLED: url:{self.url} json:{state}")

            headers = {"Content-Type": "application/json"}
            request = HTTPRequest(url=self.url,
                                  method="POST",
                                  headers=headers,
                                  body=jsonw.dumps(state),
                                  connect_timeout=self.timeout,
                                  request_timeout=self.timeout)
            for i in range(retries):
                try:
                    response = await self.client.fetch(request)
                except Exception:
                    if i == retries - 1:
                        raise
                    await asyncio.sleep(1.0)
                else:
                    break

            logging.debug(
                f"WLED: url:{self.url} status:{response.code} "
                f"response:{response.body.decode()}")

class StripSerial(Strip):
    def __init__(self: StripSerial,
                 name: str,
                 cfg: ConfigHelper):
        super().__init__(name, cfg)

        # Read the serial information (requires wled 0.13 2108250 or greater)
        self.serialport: str = cfg.get("serial")
        self.baud: int = cfg.getint("baud", 115200, above=49)

    async def send_wled_command_impl(self: StripSerial,
                                     state: Dict[str, Any]) -> None:
        async with self.request_mutex:
            if not hasattr(self, 'ser'):
                _, self.ser = await serial_asyncio.open_serial_connection(
                    url=self.serialport, baudrate=self.baud)

            logging.debug(f"WLED: serial:{self.serialport} json:{state}")

            self.ser.write(jsonw.dumps(state))

    def close(self: StripSerial):
        if hasattr(self, 'ser'):
            self.ser.close()
            logging.info(f"WLED: Closing serial {self.serialport}")

class WLED:
    def __init__(self: WLED, config: ConfigHelper) -> None:
        # root_logger = logging.getLogger()
        # root_logger.setLevel(logging.DEBUG)

        self.server = config.get_server()
        prefix_sections = config.get_prefix_sections("wled")
        logging.info(f"WLED component loading strips: {prefix_sections}")

        strip_types = {
            "HTTP": StripHttp,
            "SERIAL": StripSerial
        }
        self.strips = {}
        for section in prefix_sections:
            cfg = config[section]

            try:
                name_parts = cfg.get_name().split(maxsplit=1)
                if len(name_parts) != 2:
                    raise cfg.error(
                        f"Invalid Section Name: {cfg.get_name()}")
                name: str = name_parts[1]

                logging.info(f"WLED strip: {name}")

                # Discard old color_order setting, always support 4 color strips
                _ = cfg.get("color_order", "", deprecate=True)

                strip_type: str = cfg.get("type", "http")
                strip_class: Optional[Type[Strip]]
                strip_class = strip_types.get(strip_type.upper())
                if strip_class is None:
                    raise config.error(f"Unsupported Strip Type: {strip_type}")

                self.strips[name] = strip_class(name, cfg)

            except Exception as e:
                # Ensures errors such as "Color not supported" are visible
                msg = f"Failed to initialise strip [{cfg.get_name()}]\n{e}"
                self.server.add_warning(msg, exc_info=e)
                continue

        # Register two remote methods for GCODE
        self.server.register_remote_method(
            "set_wled_state", self.set_wled_state)
        self.server.register_remote_method(
            "set_wled", self.set_wled)

        # As moonraker is about making things a web api, let's try it
        # Yes, this is largely a cut-n-paste from power.py
        self.server.register_endpoint(
            "/machine/wled/strips", RequestType.GET, self._handle_list_strips
        )
        self.server.register_endpoint(
            "/machine/wled/status", RequestType.GET, self._handle_batch_wled_request
        )
        self.server.register_endpoint(
            "/machine/wled/on", RequestType.POST, self._handle_batch_wled_request
        )
        self.server.register_endpoint(
            "/machine/wled/off", RequestType.POST, self._handle_batch_wled_request
        )
        self.server.register_endpoint(
            "/machine/wled/toggle", RequestType.POST, self._handle_batch_wled_request
        )
        self.server.register_endpoint(
            "/machine/wled/strip", RequestType.GET | RequestType.POST,
            self._handle_single_wled_request
        )

    async def component_init(self) -> None:
        try:
            logging.debug("Initializing wled")
            event_loop = self.server.get_event_loop()
            cur_time = event_loop.get_loop_time()
            endtime = cur_time + 120.
            query_strips = list(self.strips.values())
            failed_strips: List[Strip] = []
            while cur_time < endtime:
                for strip in query_strips:
                    ret = strip.initialize()
                    if ret is not None:
                        await ret
                    if strip.error_state is not None:
                        failed_strips.append(strip)
                if not failed_strips:
                    logging.debug("All wled strips initialized")
                    return
                query_strips = failed_strips
                failed_strips = []
                await asyncio.sleep(2.)
                cur_time = event_loop.get_loop_time()
            if failed_strips:
                failed_names = [s.name for s in failed_strips]
                self.server.add_warning(
                    "The following wled strips failed init:"
                    f" {failed_names}")
        except Exception as e:
            logging.exception(e)

    async def wled_on(self: WLED, strip: str, preset: int) -> None:
        if strip not in self.strips:
            logging.info(f"Unknown WLED strip: {strip}")
            return
        await self.strips[strip].wled_on(preset)

    # Full control of wled
    # state: True, False, "on", "off"
    # preset: wled preset (int) to use (ignored if state False or "Off")
    async def set_wled_state(
        self: WLED,
        strip: str,
        state: Optional[str] = None,
        preset: int = -1,
        brightness: int = -1,
        intensity: int = -1,
        speed: int = -1
    ) -> None:
        status = None

        if isinstance(state, bool):
            status = OnOff.on if state else OnOff.off
        elif isinstance(state, str):
            status = state.lower()
            if status in ["true", "false"]:
                status = OnOff.on if status == "true" else OnOff.off

        if status is None and preset == -1 and brightness == -1 and \
           intensity == -1 and speed == -1:
            logging.info(
                "Invalid state received but no control or preset data passed"
            )
            return

        if strip not in self.strips:
            logging.info(f"Unknown WLED strip: {strip}")
            return

        # All other arguments are ignored
        if status == OnOff.off:
            await self.strips[strip].wled_off()

        # Turn on if on or a preset is specified
        if status == OnOff.on or preset != -1:
            await self.strips[strip].wled_on(preset)

        # Control
        if brightness != -1 or intensity != -1 or speed != -1:
            await self.strips[strip].wled_control(brightness, intensity, speed)

    # Individual pixel control, for compatibility with SET_LED
    async def set_wled(self: WLED,
                       strip: str,
                       red: float = 0.,
                       green: float = 0.,
                       blue: float = 0.,
                       white: float = 0.,
                       index: Optional[int] = None,
                       transmit: int = 1) -> None:
        if strip not in self.strips:
            logging.info(f"Unknown WLED strip: {strip}")
            return
        if isinstance(index, int) and index < 0:
            index = None
        await self.strips[strip].set_wled(red, green, blue, white,
                                          index,
                                          True if transmit == 1 else False)

    async def _handle_list_strips(self,
                                  web_request: WebRequest
                                  ) -> Dict[str, Any]:
        strips = {name: strip.get_strip_info()
                  for name, strip in self.strips.items()}
        output = {"strips": strips}
        return output

    async def _handle_single_wled_request(self: WLED,
                                          web_request: WebRequest
                                          ) -> Dict[str, Any]:
        strip_name: str = web_request.get_str('strip')
        preset: int = web_request.get_int('preset', -1)
        brightness: int = web_request.get_int('brightness', -1)
        intensity: int = web_request.get_int('intensity', -1)
        speed: int = web_request.get_int('speed', -1)

        req_type = web_request.get_request_type()
        if strip_name not in self.strips:
            raise self.server.error(f"No valid strip named {strip_name}")
        strip = self.strips[strip_name]
        if req_type == RequestType.GET:
            return {strip_name: strip.get_strip_info()}
        elif req_type == RequestType.POST:
            action = web_request.get_str('action').lower()
            if action not in ["on", "off", "toggle", "control"]:
                raise self.server.error(f"Invalid requested action '{action}'")
            result = await self._process_request(
                strip, action, preset, brightness, intensity, speed
            )
        return {strip_name: result}

    async def _handle_batch_wled_request(self: WLED,
                                         web_request: WebRequest
                                         ) -> Dict[str, Any]:
        args = web_request.get_args()
        ep = web_request.get_endpoint()
        if not args:
            raise self.server.error("No arguments provided")
        requested_strips = {k: self.strips.get(k, None) for k in args}
        result = {}
        req = ep.split("/")[-1]
        for name, strip in requested_strips.items():
            if strip is not None:
                result[name] = await self._process_request(strip, req, -1,
                                                           -1, -1, -1)
            else:
                result[name] = {"error": "strip_not_found"}
        return result

    async def _process_request(self: WLED,
                               strip: Strip,
                               req: str,
                               preset: int,
                               brightness: int,
                               intensity: int,
                               speed: int
                               ) -> Dict[str, Any]:
        strip_onoff = strip.onoff

        if req == "status":
            return strip.get_strip_info()
        if req == "toggle":
            req = "on" if strip_onoff == OnOff.off else "off"

        if req in ["on", "off", "control"]:
            # Always do something, could be turning off colors, or changing
            # preset, easier not to have to worry
            if req == "on" or req == "control":
                if req == "on":
                    strip_onoff = OnOff.on
                    await strip.wled_on(preset)

                if brightness != -1 or intensity != -1 or speed != -1:
                    await strip.wled_control(brightness, intensity, speed)
            else:
                strip_onoff = OnOff.off
                await strip.wled_off()

            return strip.get_strip_info()

        raise self.server.error(f"Unsupported wled request: {req}")

    def close(self) -> None:
        for strip in self.strips.values():
            strip.close()

def load_component(config: ConfigHelper) -> WLED:
    return WLED(config)



# Zip Application Deployment implementation
#
# Copyright (C) 2024  Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.

from __future__ import annotations
import pathlib
import shutil
import zipfile
import logging
from .app_deploy import AppDeploy
from .common import Channel, AppType
from ...utils import source_info
from ...utils import json_wrapper as jsonw

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Any,
    Tuple,
    Optional,
    Dict,
    List,
    Union,
    cast
)
if TYPE_CHECKING:
    from ...confighelper import ConfigHelper
    from .update_manager import CommandHelper
    from ..file_manager.file_manager import FileManager

class ZipDeploy(AppDeploy):
    def __init__(
        self,
        config: ConfigHelper,
        cmd_helper: CommandHelper
    ) -> None:
        super().__init__(config, cmd_helper, "Zip Application")
        self._configure_path(config, False)
        if self.type == AppType.ZIP:
            self._configure_virtualenv(config)
            self._configure_dependencies(config)
            self._configure_managed_services(config)
        elif self.type == AppType.WEB:
            self.prefix = f"Web Client {self.name}: "
        self.repo = config.get('repo').strip().strip("/")
        self.owner, self.project_name = self.repo.split("/", 1)
        self.persistent_files: List[str] = []
        self.warnings: List[str] = []
        self.anomalies: List[str] = []
        self.version: str = "?"
        self.remote_version: str = "?"
        self.rollback_version: str = "?"
        self.rollback_repo: str = "?"
        self.last_error: str = "?"
        self._dl_info: Tuple[str, str, int] = ("?", "?", 0)
        self._is_fallback: bool = False
        self._is_prerelease: bool = False
        self._path_writable: bool = False
        self._configure_persistent_files(config)

    def _configure_persistent_files(self, config: ConfigHelper) -> None:
        pfiles = config.getlist('persistent_files', None)
        if pfiles is not None:
            self.persistent_files = [pf.strip("/") for pf in pfiles]
            for fname in (".version", "release_info.json"):
                if fname in self.persistent_files:
                    raise config.error(
                        "Invalid value for option 'persistent_files': "
                        f"'{fname}' can not be persistent."
                    )
        if (
            self.type == AppType.ZIP and
            self.virtualenv is not None and
            self.virtualenv in self.path.parents
        ):
            rel_path = str(self.virtualenv.relative_to(self.path))
            if rel_path not in self.persistent_files:
                self.persistent_files.append(rel_path)
        if self.persistent_files:
            self.log_info(f"Configured persistent files: {self.persistent_files}")

    async def _validate_release_info(self) -> None:
        self._is_valid = False
        self._is_fallback = False
        eventloop = self.server.get_event_loop()
        self.warnings.clear()
        repo_parent = source_info.find_git_repo(self.path)
        homedir = pathlib.Path("~").expanduser()
        if not self._path_writable:
            self.warnings.append(
                f"Location at option 'path: {self.path}' is not writable."
            )
        elif not self.path.is_dir():
            self.warnings.append(
                f"Location at option 'path: {self.path}' is not a directory."
            )
        elif repo_parent is not None and repo_parent != homedir:
            self.warnings.append(
                f"Location at option 'path: {self.path}' is within a git repo. Found "
                f".git folder at '{repo_parent.joinpath('.git')}'"
            )
        else:
            rinfo = self.path.joinpath("release_info.json")
            if rinfo.is_file():
                try:
                    data = await eventloop.run_in_thread(rinfo.read_text)
                    uinfo: Dict[str, str] = jsonw.loads(data)
                    project_name = uinfo["project_name"]
                    owner = uinfo["project_owner"]
                    self.version = uinfo["version"]
                except Exception:
                    logging.exception("Failed to load release_info.json.")
                else:
                    self._is_valid = True
                    detected_repo = f"{owner}/{project_name}"
                    if self.repo.lower() != detected_repo.lower():
                        self.anomalies.append(
                            f"Value at option 'repo: {self.repo}' does not match "
                            f"detected repo '{detected_repo}', falling back to "
                            "detected version."
                        )
                        self.repo = detected_repo
                        self.owner = owner
                        self.project_name = project_name
            elif self.type == AppType.WEB:
                version_path = self.path.joinpath(".version")
                if version_path.is_file():
                    version = await eventloop.run_in_thread(version_path.read_text)
                    self.version = version.strip()
                self._is_valid = await self._detect_fallback()
        if not self._is_valid:
            self.warnings.append("Failed to validate installation")
            if self.server.is_debug_enabled():
                self.log_info("Debug Enabled, overriding validity checks")

    async def _detect_fallback(self) -> bool:
        # Only used by "web" app types to fallback on the previous version info
        fallback_defs = {
            "mainsail": "mainsail-crew",
            "fluidd": "fluidd-core"
        }
        for fname in ("manifest.json", "manifest.webmanifest"):
            manifest = self.path.joinpath(fname)
            eventloop = self.server.get_event_loop()
            if manifest.is_file():
                try:
                    mtext = await eventloop.run_in_thread(manifest.read_text)
                    mdata: Dict[str, Any] = jsonw.loads(mtext)
                    proj_name: str = mdata["name"].lower()
                except Exception:
                    self.log_exc(f"Failed to load json from {manifest}")
                    continue
                if proj_name in fallback_defs:
                    owner = fallback_defs[proj_name]
                    detected_repo = f"{owner}/{proj_name}"
                    if detected_repo != self.repo.lower():
                        self.anomalies.append(
                            f"Value at option 'repo: {self.repo}' does not match "
                            f"detected repo '{detected_repo}', falling back to "
                            "detected version."
                        )
                        self.repo = detected_repo
                        self.owner = owner
                        self.project_name = proj_name
                    self._is_fallback = True
                    return True
        return False

    async def initialize(self) -> Dict[str, Any]:
        storage = await super().initialize()
        fm: FileManager = self.server.lookup_component("file_manager")
        self._path_writable = not fm.check_reserved_path(
            self.path, need_write=True, raise_error=False
        )
        if self._path_writable and not self.path.joinpath(".writeable").is_file():
            fm.add_reserved_path(f"update_manager {self.name}", self.path)
        await self._validate_release_info()
        if self.version == "?":
            self.version = storage.get("version", "?")
        self.remote_version = storage.get('remote_version', "?")
        self.rollback_version = storage.get('rollback_version', self.version)
        self.rollback_repo = storage.get(
            'rollback_repo', self.repo if self._is_valid else "?"
        )
        self.last_error = storage.get('last_error', "")
        dl_info: List[Any] = storage.get('dl_info', ["?", "?", 0])
        self.dl_info = cast(Tuple[str, str, int], tuple(dl_info))
        if not self.needs_refresh():
            self._log_zipapp_info()
        return storage

    def get_persistent_data(self) -> Dict[str, Any]:
        storage = super().get_persistent_data()
        storage.update({
            "version": self.version,
            "remote_version": self.remote_version,
            "rollback_version": self.rollback_version,
            "rollback_repo": self.rollback_repo,
            "dl_info": list(self.dl_info),
            "last_error": self.last_error
        })
        return storage

    async def refresh(self) -> None:
        try:
            await self._validate_release_info()
            await self._get_remote_version()
        except Exception:
            logging.exception("Error Refreshing Client")
        self._log_zipapp_info()
        self._save_state()

    async def _fetch_github_version(
        self, repo: Optional[str] = None, tag: Optional[str] = None
    ) -> Dict[str, Any]:
        if repo is None:
            if not self._is_valid:
                self.log_info("Invalid Installation, aborting remote refresh")
                return {}
            repo = self.repo
        if tag is not None:
            resource = f"repos/{repo}/releases/tags/{tag}"
        elif self.channel == Channel.STABLE:
            resource = f"repos/{repo}/releases/latest"
        else:
            resource = f"repos/{repo}/releases?per_page=1"
        client = self.cmd_helper.get_http_client()
        resp = await client.github_api_request(
            resource, attempts=3, retry_pause_time=.5
        )
        release: Union[List[Any], Dict[str, Any]] = {}
        if resp.status_code == 304:
            if resp.content:
                # Not modified, however we need to restore state from
                # cached content
                release = resp.json()
            else:
                # Either not necessary or not possible to restore from cache
                return {}
        elif resp.has_error():
            self.log_info(f"Github Request Error - {resp.error}")
            self.last_error = str(resp.error)
            return {}
        else:
            release = resp.json()
        result: Dict[str, Any] = {}
        if isinstance(release, list):
            if release:
                result = release[0]
        else:
            result = release
        self.last_error = ""
        return result

    async def _get_remote_version(self) -> None:
        result = await self._fetch_github_version()
        if not result:
            return
        self.remote_version = result.get('name', "?")
        release_asset: Dict[str, Any] = result.get('assets', [{}])[0]
        dl_url: str = release_asset.get('browser_download_url', "?")
        content_type: str = release_asset.get('content_type', "?")
        size: int = release_asset.get('size', 0)
        self.dl_info = (dl_url, content_type, size)
        self._is_prerelease = result.get('prerelease', False)

    def _log_zipapp_info(self):
        warn_str = ""
        if self.warnings or self.anomalies:
            warn_str = "\nWarnings:\n"
            warn_str += "\n".join(
                [f" {item}" for item in self.warnings + self.anomalies]
            )
        dl_url, content_type, size = self.dl_info
        self.log_info(
            f"Detected\n"
            f"Repo: {self.repo}\n"
            f"Channel: {self.channel}\n"
            f"Path: {self.path}\n"
            f"Local Version: {self.version}\n"
            f"Remote Version: {self.remote_version}\n"
            f"Valid: {self._is_valid}\n"
            f"Fallback Detected: {self._is_fallback}\n"
            f"Pre-release: {self._is_prerelease}\n"
            f"Download Url: {dl_url}\n"
            f"Download Size: {size}\n"
            f"Content Type: {content_type}\n"
            f"Rollback Version: {self.rollback_version}\n"
            f"Rollback Repo: {self.rollback_repo}"
            f"{warn_str}"
        )

    def _extract_release(
        self, persist_dir: pathlib.Path, release_file: pathlib.Path
    ) -> None:
        if not persist_dir.exists():
            persist_dir.mkdir()
        if self.path.is_dir():
            # find and move persistent files
            for src_path in self.path.iterdir():
                fname = src_path.name
                if fname in self.persistent_files:
                    dest_path = persist_dir.joinpath(fname)
                    dest_dir = dest_path.parent
                    dest_dir.mkdir(parents=True, exist_ok=True)
                    shutil.move(str(src_path), str(dest_path))
            shutil.rmtree(self.path)
        self.path.mkdir()
        with zipfile.ZipFile(release_file) as zf:
            for zip_entry in zf.filelist:
                dest = pathlib.Path(zf.extract(zip_entry, str(self.path)))
                dest.chmod((zip_entry.external_attr >> 16) & 0o777)
        # Move temporary files back into
        for src_path in persist_dir.iterdir():
            dest_path = self.path.joinpath(src_path.name)
            dest_dir = dest_path.parent
            dest_dir.mkdir(parents=True, exist_ok=True)
            shutil.move(str(src_path), str(dest_path))

    async def update(
        self,
        rollback_info: Optional[Tuple[str, str, int]] = None,
        is_recover: bool = False,
        force_dep_update: bool = False
    ) -> bool:
        if not self._is_valid:
            raise self.server.error(
                f"{self.prefix}Invalid install detected, aborting update"
            )
        if rollback_info is not None:
            dl_url, content_type, size = rollback_info
            start_msg = "Rolling Back..." if not is_recover else "Recovering..."
        else:
            if self.remote_version == "?":
                await self._get_remote_version()
                if self.remote_version == "?":
                    raise self.server.error(
                        f"{self.prefix}Unable to locate update"
                    )
            dl_url, content_type, size = self.dl_info
            if self.version == self.remote_version:
                # Already up to date
                return False
            start_msg = "Updating..."
        if dl_url == "?":
            raise self.server.error(f"{self.prefix}Invalid download url")
        current_version = self.version
        event_loop = self.server.get_event_loop()
        self.notify_status(start_msg)
        self.notify_status("Downloading Release...")
        dep_info: Optional[Dict[str, Any]] = None
        if self.type == AppType.ZIP:
            dep_info = await self._collect_dependency_info()
        td = await self.cmd_helper.create_tempdir(self.name, "app")
        try:
            tempdir = pathlib.Path(td.name)
            temp_download_file = tempdir.joinpath(f"{self.name}.zip")
            temp_persist_dir = tempdir.joinpath(self.name)
            client = self.cmd_helper.get_http_client()
            await client.download_file(
                dl_url, content_type, temp_download_file, size,
                self.cmd_helper.on_download_progress
            )
            self.notify_status(
                f"Download Complete, extracting release to '{self.path}'"
            )
            await event_loop.run_in_thread(
                self._extract_release, temp_persist_dir, temp_download_file
            )
        finally:
            await event_loop.run_in_thread(td.cleanup)
        if dep_info is not None:
            await self._update_dependencies(dep_info, force_dep_update)
        self.version = self.remote_version
        await self._validate_release_info()
        if self._is_valid and rollback_info is None:
            self.rollback_version = current_version
            self.rollback_repo = self.repo
        self._log_zipapp_info()
        self._save_state()
        await self.restart_service()
        msg = "Update Finished..." if rollback_info is None else "Rollback Complete"
        self.notify_status(msg, is_complete=True)
        return True

    async def recover(
        self, hard: bool = False, force_dep_update: bool = False
    ) -> None:
        await self.update(self.dl_info, True, force_dep_update)

    async def rollback(self) -> bool:
        if self.rollback_version == "?" or self.rollback_repo == "?":
            raise self.server.error("Incomplete Rollback Data", False)
        if self.rollback_version == self.version:
            return False
        result = await self._fetch_github_version(
            self.rollback_repo, self.rollback_version
        )
        if not result:
            raise self.server.error("Failed to retrieve release asset data")
        release_asset: Dict[str, Any] = result.get('assets', [{}])[0]
        dl_url: str = release_asset.get('browser_download_url', "?")
        content_type: str = release_asset.get('content_type', "?")
        size: int = release_asset.get('size', 0)
        dl_info = (dl_url, content_type, size)
        return await self.update(dl_info)

    def get_update_status(self) -> Dict[str, Any]:
        status = super().get_update_status()
        status.update({
            'name': self.name,
            'repo_name': self.project_name,
            'owner': self.owner,
            'version': self.version,
            'remote_version': self.remote_version,
            'rollback_version': self.rollback_version,
            'last_error': self.last_error,
            'warnings': self.warnings,
            'anomalies': self.anomalies
        })
        return status
